@@ -2,7 +2,7 @@
 
   error.c -
 
-  $Author: shugo $
+  $Author: naruse $
   created at: Mon Aug  9 16:11:34 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -25,8 +25,6 @@
 #include <unistd.h>
 #endif
 
-#define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
-
 #ifndef EXIT_SUCCESS
 #define EXIT_SUCCESS 0
 #endif
@@ -39,6 +37,10 @@
 #define WEXITSTATUS(status) (status)
 #endif
 
+VALUE rb_eEAGAIN;
+VALUE rb_eEWOULDBLOCK;
+VALUE rb_eEINPROGRESS;
+
 extern const char ruby_description[];
 
 #define REPORTBUG_MSG \
@@ -73,36 +75,23 @@ err_position_0(char *buf, long len, const char *file, int line)
     }
 }
 
-static int
-err_position(char *buf, long len)
+static VALUE
+compile_snprintf(rb_encoding *enc, const char *pre, const char *file, int line, const char *fmt, va_list args)
 {
-    return err_position_0(buf, len, rb_sourcefile(), rb_sourceline());
-}
+    VALUE str = rb_enc_str_new(0, 0, enc);
 
-static void
-err_snprintf(char *buf, long len, const char *fmt, va_list args)
-{
-    long n;
-
-    n = err_position(buf, len);
-    if (len > n) {
-	vsnprintf((char*)buf+n, len-n, fmt, args);
+    if (file) {
+	rb_str_cat2(str, file);
+	if (line) rb_str_catf(str, ":%d", line);
+	rb_str_cat2(str, ": ");
     }
+    if (pre) rb_str_cat2(str, pre);
+    rb_str_vcatf(str, fmt, args);
+    return str;
 }
 
 static void
-compile_snprintf(char *buf, long len, const char *file, int line, const char *fmt, va_list args)
-{
-    long n;
-
-    n = err_position_0(buf, len, file, line);
-    if (len > n) {
-	vsnprintf((char*)buf+n, len-n, fmt, args);
-    }
-}
-
-static void
-compile_err_append(const char *s, rb_encoding *enc)
+compile_err_append(VALUE mesg)
 {
     rb_thread_t *th = GET_THREAD();
     VALUE err = th->errinfo;
@@ -112,26 +101,23 @@ compile_err_append(const char *s, rb_encoding *enc)
     /* after this line, any Ruby code *can* run */
 
     if (th->mild_compile_error) {
-	if (!RTEST(err)) {
-	    err = rb_exc_new3(rb_eSyntaxError,
-			      rb_enc_str_new(s, strlen(s), enc));
-	    th->errinfo = err;
-	}
-	else {
+	if (RTEST(err)) {
 	    VALUE str = rb_obj_as_string(err);
 
 	    rb_str_cat2(str, "\n");
-	    rb_str_cat2(str, s);
-	    th->errinfo = rb_exc_new3(rb_eSyntaxError, str);
+	    rb_str_append(str, mesg);
+	    mesg = str;
 	}
+	err = rb_exc_new3(rb_eSyntaxError, mesg);
+	th->errinfo = err;
     }
     else {
 	if (!RTEST(err)) {
 	    err = rb_exc_new2(rb_eSyntaxError, "compile error");
 	    th->errinfo = err;
 	}
-	rb_write_error(s);
-	rb_write_error("\n");
+	rb_str_cat2(mesg, "\n");
+	rb_write_error_str(mesg);
     }
 
     /* returned to the parser world */
@@ -142,62 +128,57 @@ void
 rb_compile_error_with_enc(const char *file, int line, void *enc, const char *fmt, ...)
 {
     va_list args;
-    char buf[BUFSIZ];
+    VALUE str;
 
     va_start(args, fmt);
-    compile_snprintf(buf, BUFSIZ, file, line, fmt, args);
+    str = compile_snprintf(enc, NULL, file, line, fmt, args);
     va_end(args);
-    compile_err_append(buf, (rb_encoding *)enc);
+    compile_err_append(str);
 }
 
 void
 rb_compile_error(const char *file, int line, const char *fmt, ...)
 {
     va_list args;
-    char buf[BUFSIZ];
+    VALUE str;
 
     va_start(args, fmt);
-    compile_snprintf(buf, BUFSIZ, file, line, fmt, args);
+    str = compile_snprintf(NULL, NULL, file, line, fmt, args);
     va_end(args);
-    compile_err_append(buf, NULL);
+    compile_err_append(str);
 }
 
 void
 rb_compile_error_append(const char *fmt, ...)
 {
     va_list args;
-    char buf[BUFSIZ];
+    VALUE str;
 
     va_start(args, fmt);
-    vsnprintf(buf, BUFSIZ, fmt, args);
+    str = rb_vsprintf(fmt, args);
     va_end(args);
-    compile_err_append(buf, NULL);
+    compile_err_append(str);
 }
 
 static void
 compile_warn_print(const char *file, int line, const char *fmt, va_list args)
 {
-    char buf[BUFSIZ];
-    int len;
+    VALUE str;
 
-    compile_snprintf(buf, BUFSIZ, file, line, fmt, args);
-    len = (int)strlen(buf);
-    buf[len++] = '\n';
-    rb_write_error2(buf, len);
+    str = compile_snprintf(NULL, "warning: ", file, line, fmt, args);
+    rb_str_cat2(str, "\n");
+    rb_write_error_str(str);
 }
 
 void
 rb_compile_warn(const char *file, int line, const char *fmt, ...)
 {
-    char buf[BUFSIZ];
     va_list args;
 
     if (NIL_P(ruby_verbose)) return;
 
-    snprintf(buf, BUFSIZ, "warning: %s", fmt);
-
     va_start(args, fmt);
-    compile_warn_print(file, line, buf, args);
+    compile_warn_print(file, line, fmt, args);
     va_end(args);
 }
 
@@ -205,42 +186,43 @@ rb_compile_warn(const char *file, int line, const char *fmt, ...)
 void
 rb_compile_warning(const char *file, int line, const char *fmt, ...)
 {
-    char buf[BUFSIZ];
     va_list args;
 
     if (!RTEST(ruby_verbose)) return;
 
-    snprintf(buf, BUFSIZ, "warning: %s", fmt);
-
     va_start(args, fmt);
-    compile_warn_print(file, line, buf, args);
+    compile_warn_print(file, line, fmt, args);
     va_end(args);
 }
 
 static void
 warn_print(const char *fmt, va_list args)
 {
-    char buf[BUFSIZ];
-    int len;
+    VALUE str = rb_str_new(0, 0);
+    VALUE file = rb_sourcefilename();
 
-    err_snprintf(buf, BUFSIZ, fmt, args);
-    len = (int)strlen(buf);
-    buf[len++] = '\n';
-    rb_write_error2(buf, len);
+    if (!NIL_P(file)) {
+	int line = rb_sourceline();
+	str = rb_str_append(str, file);
+	if (line) rb_str_catf(str, ":%d", line);
+	rb_str_cat2(str, ": ");
+    }
+
+    rb_str_cat2(str, "warning: ");
+    rb_str_vcatf(str, fmt, args);
+    rb_str_cat2(str, "\n");
+    rb_write_error_str(str);
 }
 
 void
 rb_warn(const char *fmt, ...)
 {
-    char buf[BUFSIZ];
     va_list args;
 
     if (NIL_P(ruby_verbose)) return;
 
-    snprintf(buf, BUFSIZ, "warning: %s", fmt);
-
     va_start(args, fmt);
-    warn_print(buf, args);
+    warn_print(fmt, args);
     va_end(args);
 }
 
@@ -248,15 +230,12 @@ rb_warn(const char *fmt, ...)
 void
 rb_warning(const char *fmt, ...)
 {
-    char buf[BUFSIZ];
     va_list args;
 
     if (!RTEST(ruby_verbose)) return;
 
-    snprintf(buf, BUFSIZ, "warning: %s", fmt);
-
     va_start(args, fmt);
-    warn_print(buf, args);
+    warn_print(fmt, args);
     va_end(args);
 }
 
@@ -432,8 +411,9 @@ rb_builtin_type_name(int t)
     return 0;
 }
 
-static const char *
-builtin_class_name(VALUE x)
+#define builtin_class_name rb_builtin_class_name
+const char *
+rb_builtin_class_name(VALUE x)
 {
     const char *etype;
 
@@ -552,7 +532,7 @@ VALUE rb_eSystemCallError;
 VALUE rb_mErrno;
 static VALUE rb_eNOERROR;
 
-#undef rb_exc_new2
+#undef rb_exc_new_cstr
 
 VALUE
 rb_exc_new(VALUE etype, const char *ptr, long len)
@@ -561,13 +541,13 @@ rb_exc_new(VALUE etype, const char *ptr, long len)
 }
 
 VALUE
-rb_exc_new2(VALUE etype, const char *s)
+rb_exc_new_cstr(VALUE etype, const char *s)
 {
     return rb_exc_new(etype, s, strlen(s));
 }
 
 VALUE
-rb_exc_new3(VALUE etype, VALUE str)
+rb_exc_new_str(VALUE etype, VALUE str)
 {
     StringValue(str);
     return rb_funcall(etype, rb_intern("new"), 1, str);
@@ -741,7 +721,8 @@ rb_check_backtrace(VALUE bt)
 	    rb_raise(rb_eTypeError, err);
 	}
 	for (i=0;i<RARRAY_LEN(bt);i++) {
-	    if (!RB_TYPE_P(RARRAY_PTR(bt)[i], T_STRING)) {
+	    VALUE e = RARRAY_AREF(bt, i);
+	    if (!RB_TYPE_P(e, T_STRING)) {
 		rb_raise(rb_eTypeError, err);
 	    }
 	}
@@ -1119,7 +1100,7 @@ name_err_mesg_to_str(VALUE obj)
 	if (desc && desc[0] != '#') {
 	    d = d ? rb_str_dup(d) : rb_str_new2(desc);
 	    rb_str_cat2(d, ":");
-	    rb_str_cat2(d, rb_obj_classname(obj));
+	    rb_str_append(d, rb_class_name(CLASS_OF(obj)));
 	}
 	args[0] = mesg;
 	args[1] = ptr[2];
@@ -1129,6 +1110,13 @@ name_err_mesg_to_str(VALUE obj)
     return mesg;
 }
 
+/* :nodoc: */
+static VALUE
+name_err_mesg_dump(VALUE obj, VALUE limit)
+{
+    return name_err_mesg_to_str(obj);
+}
+
 /* :nodoc: */
 static VALUE
 name_err_mesg_load(VALUE klass, VALUE str)
@@ -1198,6 +1186,24 @@ set_syserr(int n, const char *name)
 
     if (!st_lookup(syserr_tbl, n, &error)) {
 	error = rb_define_class_under(rb_mErrno, name, rb_eSystemCallError);
+
+	/* capture nonblock errnos for WaitReadable/WaitWritable subclasses */
+	switch (n) {
+	    case EAGAIN:
+	        rb_eEAGAIN = error;
+
+#if EAGAIN != EWOULDBLOCK
+                break;
+	    case EWOULDBLOCK:
+#endif
+
+		rb_eEWOULDBLOCK = error;
+		break;
+	    case EINPROGRESS:
+		rb_eEINPROGRESS = error;
+		break;
+	}
+
 	rb_define_const(error, "Errno", INT2NUM(n));
 	st_add_direct(syserr_tbl, n, error);
     }
@@ -1239,12 +1245,12 @@ syserr_initialize(int argc, VALUE *argv, VALUE self)
     char *strerror();
 #endif
     const char *err;
-    VALUE mesg, error;
+    VALUE mesg, error, func;
     VALUE klass = rb_obj_class(self);
 
     if (klass == rb_eSystemCallError) {
 	st_data_t data = (st_data_t)klass;
-	rb_scan_args(argc, argv, "11", &mesg, &error);
+	rb_scan_args(argc, argv, "12", &mesg, &error, &func);
 	if (argc == 1 && FIXNUM_P(mesg)) {
 	    error = mesg; mesg = Qnil;
 	}
@@ -1254,11 +1260,11 @@ syserr_initialize(int argc, VALUE *argv, VALUE self)
 	    if (!RB_TYPE_P(self, T_OBJECT)) { /* insurance to avoid type crash */
 		rb_raise(rb_eTypeError, "invalid instance type");
 	    }
-	    RBASIC(self)->klass = klass;
+	    RBASIC_SET_CLASS(self, klass);
 	}
     }
     else {
-	rb_scan_args(argc, argv, "01", &mesg);
+	rb_scan_args(argc, argv, "02", &mesg, &func);
 	error = rb_const_get(klass, rb_intern("Errno"));
     }
     if (!NIL_P(error)) err = strerror(NUM2INT(error));
@@ -1268,8 +1274,10 @@ syserr_initialize(int argc, VALUE *argv, VALUE self)
 	VALUE str = StringValue(mesg);
 	rb_encoding *me = rb_enc_get(mesg);
 
-	mesg = rb_sprintf("%s - %.*s", err,
-			  (int)RSTRING_LEN(str), RSTRING_PTR(str));
+	if (NIL_P(func))
+	    mesg = rb_sprintf("%s - %"PRIsVALUE, err, mesg);
+	else
+	    mesg = rb_sprintf("%s @ %"PRIsVALUE" - %"PRIsVALUE, err, func, mesg);
 	if (le != me && rb_enc_asciicompat(me)) {
 	    le = me;
 	}/* else assume err is non ASCII string. */
@@ -1365,6 +1373,7 @@ syserr_eqq(VALUE self, VALUE exc)
  *
  *     begin
  *       Process.kill('HUP',Process.pid)
+ *       sleep # wait for receiver to handle signal sent by Process.kill
  *     rescue SignalException => e
  *       puts "received Exception #{e}"
  *     end
@@ -1406,7 +1415,7 @@ syserr_eqq(VALUE self, VALUE exc)
  *
  *  <em>raises the exception:</em>
  *
- *     TypeError: can't convert String into Integer
+ *     TypeError: no implicit conversion of String into Integer
  *
  */
 
@@ -1579,14 +1588,14 @@ syserr_eqq(VALUE self, VALUE exc)
  *
  *     foo = "bar"
  *     proc = Proc.new do
- *       $SAFE = 4
- *       foo.gsub! "a", "*"
+ *       $SAFE = 3
+ *       foo.untaint
  *     end
  *     proc.call
  *
  *  <em>raises the exception:</em>
  *
- *     SecurityError: Insecure: can't modify string
+ *     SecurityError: Insecure: Insecure operation `untaint' at level 3
  */
 
 /*
@@ -1743,6 +1752,7 @@ Init_Exception(void)
     rb_eSyntaxError = rb_define_class("SyntaxError", rb_eScriptError);
 
     rb_eLoadError   = rb_define_class("LoadError", rb_eScriptError);
+    /* the path failed to load */
     rb_attr(rb_eLoadError, rb_intern("path"), 1, 0, Qfalse);
 
     rb_eNotImpError = rb_define_class("NotImplementedError", rb_eScriptError);
@@ -1755,7 +1765,7 @@ Init_Exception(void)
     rb_define_singleton_method(rb_cNameErrorMesg, "!", rb_name_err_mesg_new, NAME_ERR_MESG_COUNT);
     rb_define_method(rb_cNameErrorMesg, "==", name_err_mesg_equal, 1);
     rb_define_method(rb_cNameErrorMesg, "to_str", name_err_mesg_to_str, 0);
-    rb_define_method(rb_cNameErrorMesg, "_dump", name_err_mesg_to_str, 1);
+    rb_define_method(rb_cNameErrorMesg, "_dump", name_err_mesg_dump, 1);
     rb_define_singleton_method(rb_cNameErrorMesg, "_load", name_err_mesg_load, 1);
     rb_eNoMethodError = rb_define_class("NoMethodError", rb_eNameError);
     rb_define_method(rb_eNoMethodError, "initialize", nometh_err_initialize, -1);
@@ -1922,6 +1932,27 @@ rb_sys_fail_str(VALUE mesg)
     rb_exc_raise(make_errno_exc_str(mesg));
 }
 
+#ifdef RUBY_FUNCTION_NAME_STRING
+void
+rb_sys_fail_path_in(const char *func_name, VALUE path)
+{
+    int n = errno;
+    VALUE args[2];
+
+    errno = 0;
+    if (!path) path = Qnil;
+    if (n == 0) {
+	const char *s = !NIL_P(path) ? RSTRING_PTR(path) : "";
+	if (!func_name) func_name = "(null)";
+	rb_bug("rb_sys_fail_path_in(%s, %s) - errno == 0",
+	       func_name, s);
+    }
+    args[0] = path;
+    args[1] = rb_str_new_cstr(func_name);
+    rb_exc_raise(rb_class_new_instance(2, args, get_syserr(n)));
+}
+#endif
+
 void
 rb_mod_sys_fail(VALUE mod, const char *mesg)
 {
@@ -1999,17 +2030,12 @@ rb_check_frozen(VALUE obj)
 void
 rb_error_untrusted(VALUE obj)
 {
-    if (rb_safe_level() >= 4) {
-	rb_raise(rb_eSecurityError, "Insecure: can't modify %s",
-		 rb_obj_classname(obj));
-    }
 }
 
 #undef rb_check_trusted
 void
 rb_check_trusted(VALUE obj)
 {
-    rb_check_trusted_internal(obj);
 }
 
 void
@@ -2017,9 +2043,8 @@ rb_check_copyable(VALUE obj, VALUE orig)
 {
     if (!FL_ABLE(obj)) return;
     rb_check_frozen_internal(obj);
-    rb_check_trusted_internal(obj);
     if (!FL_ABLE(orig)) return;
-    if ((~RBASIC(obj)->flags & RBASIC(orig)->flags) & (FL_UNTRUSTED|FL_TAINT)) {
+    if ((~RBASIC(obj)->flags & RBASIC(orig)->flags) & FL_TAINT) {
 	if (rb_safe_level() > 0) {
 	    rb_raise(rb_eSecurityError, "Insecure: can't modify %"PRIsVALUE,
 		     RBASIC(obj)->klass);
