@@ -2,7 +2,7 @@
 
   stringio.c -
 
-  $Author: matz $
+  $Author: ko1 $
   $RoughId: stringio.c,v 1.13 2002/03/14 03:24:18 nobu Exp $
   created at: Tue Feb 19 04:10:38 JST 2002
 
@@ -28,15 +28,13 @@ struct StringIO {
     int count;
 };
 
-static void strio_mark _((struct StringIO *));
-static void strio_free _((struct StringIO *));
-static void strio_init(int, VALUE *, struct StringIO *);
+static void strio_init(int, VALUE *, struct StringIO *, VALUE);
 
-#define IS_STRIO(obj) (RDATA(obj)->dmark == (RUBY_DATA_FUNC)strio_mark)
+#define IS_STRIO(obj) (rb_typeddata_is_kind_of((obj), &strio_data_type))
 #define error_inval(msg) (errno = EINVAL, rb_sys_fail(msg))
 
 static struct StringIO *
-strio_alloc()
+strio_alloc(void)
 {
     struct StringIO *ptr = ALLOC(struct StringIO);
     ptr->string = Qnil;
@@ -48,36 +46,46 @@ strio_alloc()
 }
 
 static void
-strio_mark(struct StringIO *ptr)
+strio_mark(void *p)
 {
+    struct StringIO *ptr = p;
     if (ptr) {
 	rb_gc_mark(ptr->string);
     }
 }
 
 static void
-strio_free(struct StringIO *ptr)
+strio_free(void *p)
 {
+    struct StringIO *ptr = p;
     if (--ptr->count <= 0) {
 	xfree(ptr);
     }
 }
 
-static struct StringIO*
-check_strio(VALUE self)
+static size_t
+strio_memsize(const void *p)
 {
-    Check_Type(self, T_DATA);
-    if (!IS_STRIO(self)) {
-	rb_raise(rb_eTypeError, "wrong argument type %s (expected StringIO)",
-		 rb_class2name(CLASS_OF(self)));
-    }
-    return DATA_PTR(self);
+    const struct StringIO *ptr = p;
+    if (!ptr) return 0;
+    return sizeof(struct StringIO);
 }
 
+static const rb_data_type_t strio_data_type = {
+    "strio",
+    {
+	strio_mark,
+	strio_free,
+	strio_memsize,
+    },
+};
+
+#define check_strio(self) ((struct StringIO*)rb_check_typeddata((self), &strio_data_type))
+
 static struct StringIO*
 get_strio(VALUE self)
 {
-    struct StringIO *ptr = check_strio(self);
+    struct StringIO *ptr = check_strio(rb_io_taint_check(self));
 
     if (!ptr) {
 	rb_raise(rb_eIOError, "uninitialized stream");
@@ -86,40 +94,51 @@ get_strio(VALUE self)
 }
 
 static VALUE
-strio_substr(struct StringIO *ptr, int pos, int len)
+strio_substr(struct StringIO *ptr, long pos, long len)
 {
     VALUE str = ptr->string;
     rb_encoding *enc = rb_enc_get(str);
-    int rlen = RSTRING_LEN(str) - pos;
+    long rlen = RSTRING_LEN(str) - pos;
 
     if (len > rlen) len = rlen;
     if (len < 0) len = 0;
+    if (len == 0) return rb_str_new(0,0);
     return rb_enc_str_new(RSTRING_PTR(str)+pos, len, enc);
 }
 
 #define StringIO(obj) get_strio(obj)
 
-#define CLOSED(ptr) (!((ptr)->flags & FMODE_READWRITE))
-#define READABLE(ptr) ((ptr)->flags & FMODE_READABLE)
-#define WRITABLE(ptr) ((ptr)->flags & FMODE_WRITABLE)
+#define STRIO_READABLE FL_USER4
+#define STRIO_WRITABLE FL_USER5
+#define STRIO_READWRITE (STRIO_READABLE|STRIO_WRITABLE)
+typedef char strio_flags_check[(STRIO_READABLE/FMODE_READABLE == STRIO_WRITABLE/FMODE_WRITABLE) * 2 - 1];
+#define STRIO_MODE_SET_P(strio, mode) \
+    ((RBASIC(strio)->flags & STRIO_##mode) && \
+     ((struct StringIO*)DATA_PTR(strio))->flags & FMODE_##mode)
+#define CLOSED(strio) (!STRIO_MODE_SET_P(strio, READWRITE))
+#define READABLE(strio) STRIO_MODE_SET_P(strio, READABLE)
+#define WRITABLE(strio) STRIO_MODE_SET_P(strio, WRITABLE)
+
+static VALUE sym_exception;
 
 static struct StringIO*
-readable(struct StringIO *ptr)
+readable(VALUE strio)
 {
-    if (!READABLE(ptr)) {
+    struct StringIO *ptr = StringIO(strio);
+    if (!READABLE(strio)) {
 	rb_raise(rb_eIOError, "not opened for reading");
     }
     return ptr;
 }
 
 static struct StringIO*
-writable(struct StringIO *ptr)
+writable(VALUE strio)
 {
-    if (!WRITABLE(ptr)) {
+    struct StringIO *ptr = StringIO(strio);
+    if (!WRITABLE(strio)) {
 	rb_raise(rb_eIOError, "not opened for writing");
     }
     if (!OBJ_TAINTED(ptr->string)) {
-	rb_secure(4);
     }
     return ptr;
 }
@@ -135,7 +154,7 @@ check_modifiable(struct StringIO *ptr)
 static VALUE
 strio_s_allocate(VALUE klass)
 {
-    return Data_Wrap_Struct(klass, strio_mark, strio_free, 0);
+    return TypedData_Wrap_Struct(klass, &strio_data_type, 0);
 }
 
 /*
@@ -152,15 +171,15 @@ strio_initialize(int argc, VALUE *argv, VALUE self)
 	DATA_PTR(self) = ptr = strio_alloc();
     }
     rb_call_super(0, 0);
-    strio_init(argc, argv, ptr);
+    strio_init(argc, argv, ptr, self);
     return self;
 }
 
 static void
-strio_init(int argc, VALUE *argv, struct StringIO *ptr)
+strio_init(int argc, VALUE *argv, struct StringIO *ptr, VALUE self)
 {
     VALUE string, mode;
-    int trunc = Qfalse;
+    int trunc = 0;
 
     switch (rb_scan_args(argc, argv, "02", &string, &mode)) {
       case 2:
@@ -195,6 +214,7 @@ strio_init(int argc, VALUE *argv, struct StringIO *ptr)
     ptr->string = string;
     ptr->pos = 0;
     ptr->lineno = 0;
+    RBASIC(self)->flags |= (ptr->flags & FMODE_READWRITE) * (STRIO_READABLE / FMODE_READABLE);
 }
 
 static VALUE
@@ -279,7 +299,8 @@ strio_unimpl(int argc, VALUE *argv, VALUE self)
 {
     StringIO(self);
     rb_notimplement();
-    return Qnil;		/* not reached */
+
+    UNREACHABLE;
 }
 
 /*
@@ -304,7 +325,7 @@ strio_set_string(VALUE self, VALUE string)
 {
     struct StringIO *ptr = StringIO(self);
 
-    if (!OBJ_TAINTED(self)) rb_secure(4);
+    rb_io_taint_check(self);
     ptr->flags &= ~FMODE_READWRITE;
     StringValue(string);
     ptr->flags = OBJ_FROZEN(string) ? FMODE_READABLE : FMODE_READWRITE;
@@ -317,17 +338,17 @@ strio_set_string(VALUE self, VALUE string)
  * call-seq:
  *   strio.close  -> nil
  *
- * Closes strio.  The *strio* is unavailable for any further data 
+ * Closes strio.  The *strio* is unavailable for any further data
  * operations; an +IOError+ is raised if such an attempt is made.
  */
 static VALUE
 strio_close(VALUE self)
 {
-    struct StringIO *ptr = StringIO(self);
-    if (CLOSED(ptr)) {
+    StringIO(self);
+    if (CLOSED(self)) {
 	rb_raise(rb_eIOError, "closed stream");
     }
-    ptr->flags &= ~FMODE_READWRITE;
+    RBASIC(self)->flags &= ~STRIO_READWRITE;
     return Qnil;
 }
 
@@ -341,11 +362,11 @@ strio_close(VALUE self)
 static VALUE
 strio_close_read(VALUE self)
 {
-    struct StringIO *ptr = StringIO(self);
-    if (!READABLE(ptr)) {
+    StringIO(self);
+    if (!READABLE(self)) {
 	rb_raise(rb_eIOError, "closing non-duplex IO for reading");
     }
-    ptr->flags &= ~FMODE_READABLE;
+    RBASIC(self)->flags &= ~STRIO_READABLE;
     return Qnil;
 }
 
@@ -359,11 +380,11 @@ strio_close_read(VALUE self)
 static VALUE
 strio_close_write(VALUE self)
 {
-    struct StringIO *ptr = StringIO(self);
-    if (!WRITABLE(ptr)) {
+    StringIO(self);
+    if (!WRITABLE(self)) {
 	rb_raise(rb_eIOError, "closing non-duplex IO for writing");
     }
-    ptr->flags &= ~FMODE_WRITABLE;
+    RBASIC(self)->flags &= ~STRIO_WRITABLE;
     return Qnil;
 }
 
@@ -376,8 +397,8 @@ strio_close_write(VALUE self)
 static VALUE
 strio_closed(VALUE self)
 {
-    struct StringIO *ptr = StringIO(self);
-    if (!CLOSED(ptr)) return Qfalse;
+    StringIO(self);
+    if (!CLOSED(self)) return Qfalse;
     return Qtrue;
 }
 
@@ -390,8 +411,8 @@ strio_closed(VALUE self)
 static VALUE
 strio_closed_read(VALUE self)
 {
-    struct StringIO *ptr = StringIO(self);
-    if (READABLE(ptr)) return Qfalse;
+    StringIO(self);
+    if (READABLE(self)) return Qfalse;
     return Qtrue;
 }
 
@@ -404,8 +425,8 @@ strio_closed_read(VALUE self)
 static VALUE
 strio_closed_write(VALUE self)
 {
-    struct StringIO *ptr = StringIO(self);
-    if (WRITABLE(ptr)) return Qfalse;
+    StringIO(self);
+    if (WRITABLE(self)) return Qfalse;
     return Qtrue;
 }
 
@@ -414,13 +435,13 @@ strio_closed_write(VALUE self)
  *   strio.eof     -> true or false
  *   strio.eof?    -> true or false
  *
- * Returns true if *strio* is at end of file. The stringio must be  
+ * Returns true if *strio* is at end of file. The stringio must be
  * opened for reading or an +IOError+ will be raised.
  */
 static VALUE
 strio_eof(VALUE self)
 {
-    struct StringIO *ptr = readable(StringIO(self));
+    struct StringIO *ptr = readable(self);
     if (ptr->pos < RSTRING_LEN(ptr->string)) return Qfalse;
     return Qtrue;
 }
@@ -439,6 +460,8 @@ strio_copy(VALUE copy, VALUE orig)
     }
     DATA_PTR(copy) = ptr;
     OBJ_INFECT(copy, orig);
+    RBASIC(copy)->flags &= ~STRIO_READWRITE;
+    RBASIC(copy)->flags |= RBASIC(orig)->flags & STRIO_READWRITE;
     ++ptr->count;
     return copy;
 }
@@ -473,16 +496,12 @@ strio_set_lineno(VALUE self, VALUE lineno)
     return lineno;
 }
 
-/* call-seq: strio.binmode -> true */
 #define strio_binmode strio_self
 
-/* call-seq: strio.fcntl */
 #define strio_fcntl strio_unimpl
 
-/* call-seq: strio.flush -> strio */
 #define strio_flush strio_self
 
-/* call-seq: strio.fsync -> 0 */
 #define strio_fsync strio_0
 
 /*
@@ -490,17 +509,17 @@ strio_set_lineno(VALUE self, VALUE lineno)
  *   strio.reopen(other_StrIO)     -> strio
  *   strio.reopen(string, mode)    -> strio
  *
- * Reinitializes *strio* with the given <i>other_StrIO</i> or _string_ 
+ * Reinitializes *strio* with the given <i>other_StrIO</i> or _string_
  * and _mode_ (see StringIO#new).
  */
 static VALUE
 strio_reopen(int argc, VALUE *argv, VALUE self)
 {
-    if (!OBJ_TAINTED(self)) rb_secure(4);
-    if (argc == 1 && TYPE(*argv) != T_STRING) {
+    rb_io_taint_check(self);
+    if (argc == 1 && !RB_TYPE_P(*argv, T_STRING)) {
 	return strio_copy(self, *argv);
     }
-    strio_init(argc, argv, StringIO(self));
+    strio_init(argc, argv, StringIO(self), self);
     return self;
 }
 
@@ -567,7 +586,7 @@ strio_seek(int argc, VALUE *argv, VALUE self)
 
     rb_scan_args(argc, argv, "11", NULL, &whence);
     offset = NUM2LONG(argv[0]);
-    if (CLOSED(ptr)) {
+    if (CLOSED(self)) {
 	rb_raise(rb_eIOError, "closed stream");
     }
     switch (NIL_P(whence) ? 0 : NUM2LONG(whence)) {
@@ -602,7 +621,6 @@ strio_get_sync(VALUE self)
     return Qtrue;
 }
 
-/* call-seq: strio.sync = boolean -> boolean */
 #define strio_set_sync strio_first
 
 #define strio_tell strio_get_pos
@@ -610,13 +628,14 @@ strio_get_sync(VALUE self)
 /*
  * call-seq:
  *   strio.each_byte {|byte| block }  -> strio
+ *   strio.each_byte                  -> anEnumerator
  *
  * See IO#each_byte.
  */
 static VALUE
 strio_each_byte(VALUE self)
 {
-    struct StringIO *ptr = readable(StringIO(self));
+    struct StringIO *ptr = readable(self);
 
     RETURN_ENUMERATOR(self, 0, 0);
 
@@ -627,6 +646,18 @@ strio_each_byte(VALUE self)
     return self;
 }
 
+/*
+ *  This is a deprecated alias for #each_byte.
+ */
+static VALUE
+strio_bytes(VALUE self)
+{
+    rb_warn("StringIO#bytes is deprecated; use #each_byte instead");
+    if (!rb_block_given_p())
+	return rb_enumeratorize(self, ID2SYM(rb_intern("each_byte")), 0, 0);
+    return strio_each_byte(self);
+}
+
 /*
  * call-seq:
  *   strio.getc   -> string or nil
@@ -636,7 +667,7 @@ strio_each_byte(VALUE self)
 static VALUE
 strio_getc(VALUE self)
 {
-    struct StringIO *ptr = readable(StringIO(self));
+    struct StringIO *ptr = readable(self);
     rb_encoding *enc = rb_enc_get(ptr->string);
     int len;
     char *p;
@@ -659,7 +690,7 @@ strio_getc(VALUE self)
 static VALUE
 strio_getbyte(VALUE self)
 {
-    struct StringIO *ptr = readable(StringIO(self));
+    struct StringIO *ptr = readable(self);
     int c;
     if (ptr->pos >= RSTRING_LEN(ptr->string)) {
 	return Qnil;
@@ -690,20 +721,20 @@ strio_extend(struct StringIO *ptr, long pos, long len)
  *   strio.ungetc(string)   -> nil
  *
  * Pushes back one character (passed as a parameter) onto *strio*
- * such that a subsequent buffered read will return it.  Pushing back 
- * behind the beginning of the buffer string is not possible.  Nothing
- * will be done if such an attempt is made.
- * In other case, there is no limitation for multiple pushbacks.
+ * such that a subsequent buffered read will return it.  There is no
+ * limitation for multiple pushbacks including pushing back behind the
+ * beginning of the buffer string.
  */
 static VALUE
 strio_ungetc(VALUE self, VALUE c)
 {
-    struct StringIO *ptr = readable(StringIO(self));
+    struct StringIO *ptr = readable(self);
     long lpos, clen;
     char *p, *pend;
     rb_encoding *enc, *enc2;
 
     if (NIL_P(c)) return Qnil;
+    check_modifiable(ptr);
     if (FIXNUM_P(c)) {
 	int cc = FIX2INT(c);
 	char buf[16];
@@ -720,16 +751,26 @@ strio_ungetc(VALUE self, VALUE c)
 	    c = rb_str_conv_enc(c, enc2, enc);
 	}
     }
-    /* get logical position */
-    lpos = 0; p = RSTRING_PTR(ptr->string); pend = p + ptr->pos - 1;
-    for (;;) {
-	clen = rb_enc_mbclen(p, pend, enc);
-	if (p+clen >= pend) break;
-	p += clen;
-	lpos++;
+    if (RSTRING_LEN(ptr->string) < ptr->pos) {
+	long len = RSTRING_LEN(ptr->string);
+	rb_str_resize(ptr->string, ptr->pos - 1);
+	memset(RSTRING_PTR(ptr->string) + len, 0, ptr->pos - len - 1);
+	rb_str_concat(ptr->string, c);
+	ptr->pos--;
+    }
+    else {
+	/* get logical position */
+	lpos = 0; p = RSTRING_PTR(ptr->string); pend = p + ptr->pos;
+	for (;;) {
+	    clen = rb_enc_mbclen(p, pend, enc);
+	    if (p+clen >= pend) break;
+	    p += clen;
+	    lpos++;
+	}
+	clen = p - RSTRING_PTR(ptr->string);
+	rb_str_update(ptr->string, lpos, ptr->pos ? 1 : 0, c);
+	ptr->pos = clen;
     }
-    rb_str_update(ptr->string, lpos, ptr->pos ? 1 : 0, c);
-    ptr->pos = p - RSTRING_PTR(ptr->string);
 
     return Qnil;
 }
@@ -743,20 +784,50 @@ strio_ungetc(VALUE self, VALUE c)
 static VALUE
 strio_ungetbyte(VALUE self, VALUE c)
 {
-    NUM2INT(c);
-    return strio_ungetc(self, c);
+    struct StringIO *ptr = readable(self);
+    char buf[1], *cp = buf;
+    long pos = ptr->pos, cl = 1;
+    VALUE str = ptr->string;
+
+    if (NIL_P(c)) return Qnil;
+    if (FIXNUM_P(c)) {
+	buf[0] = (char)FIX2INT(c);
+    }
+    else {
+	SafeStringValue(c);
+	cp = RSTRING_PTR(c);
+	cl = RSTRING_LEN(c);
+	if (cl == 0) return Qnil;
+    }
+    check_modifiable(ptr);
+    rb_str_modify(str);
+    if (cl > pos) {
+	char *s;
+	long rest = RSTRING_LEN(str) - pos;
+	rb_str_resize(str, rest + cl);
+	s = RSTRING_PTR(str);
+	memmove(s + cl, s + pos, rest);
+	pos = 0;
+    }
+    else {
+	pos -= cl;
+    }
+    memcpy(RSTRING_PTR(str) + pos, cp, cl);
+    ptr->pos = pos;
+    RB_GC_GUARD(c);
+    return Qnil;
 }
 
 /*
  * call-seq:
- *   strio.readchar   -> fixnum
+ *   strio.readchar   -> string
  *
  * See IO#readchar.
  */
 static VALUE
 strio_readchar(VALUE self)
 {
-    VALUE c = strio_getc(self);
+    VALUE c = rb_funcall2(self, rb_intern("getc"), 0, 0);
     if (NIL_P(c)) rb_eof_error();
     return c;
 }
@@ -770,7 +841,7 @@ strio_readchar(VALUE self)
 static VALUE
 strio_readbyte(VALUE self)
 {
-    VALUE c = strio_getbyte(self);
+    VALUE c = rb_funcall2(self, rb_intern("getbyte"), 0, 0);
     if (NIL_P(c)) rb_eof_error();
     return c;
 }
@@ -778,6 +849,7 @@ strio_readbyte(VALUE self)
 /*
  * call-seq:
  *   strio.each_char {|char| block }  -> strio
+ *   strio.each_char                  -> anEnumerator
  *
  * See IO#each_char.
  */
@@ -794,6 +866,62 @@ strio_each_char(VALUE self)
     return self;
 }
 
+/*
+ *  This is a deprecated alias for <code>each_char</code>.
+ */
+static VALUE
+strio_chars(VALUE self)
+{
+    rb_warn("StringIO#chars is deprecated; use #each_char instead");
+    if (!rb_block_given_p())
+	return rb_enumeratorize(self, ID2SYM(rb_intern("each_char")), 0, 0);
+    return strio_each_char(self);
+}
+
+/*
+ * call-seq:
+ *   strio.each_codepoint {|c| block }  -> strio
+ *   strio.each_codepoint               -> anEnumerator
+ *
+ * See IO#each_codepoint.
+ */
+static VALUE
+strio_each_codepoint(VALUE self)
+{
+    struct StringIO *ptr;
+    rb_encoding *enc;
+    unsigned int c;
+    int n;
+
+    RETURN_ENUMERATOR(self, 0, 0);
+
+    ptr = readable(self);
+    enc = rb_enc_get(ptr->string);
+    for (;;) {
+	if (ptr->pos >= RSTRING_LEN(ptr->string)) {
+	    return self;
+	}
+
+	c = rb_enc_codepoint_len(RSTRING_PTR(ptr->string)+ptr->pos,
+				 RSTRING_END(ptr->string), &n, enc);
+	rb_yield(UINT2NUM(c));
+	ptr->pos += n;
+    }
+    return self;
+}
+
+/*
+ *  This is a deprecated alias for <code>each_codepoint</code>.
+ */
+static VALUE
+strio_codepoints(VALUE self)
+{
+    rb_warn("StringIO#codepoints is deprecated; use #each_codepoint instead");
+    if (!rb_block_given_p())
+	return rb_enumeratorize(self, ID2SYM(rb_intern("each_codepoint")), 0, 0);
+    return strio_each_codepoint(self);
+}
+
 /* Boyer-Moore search: copied from regex.c */
 static void
 bm_init_skip(long *skip, const char *pat, long m)
@@ -832,18 +960,17 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
 {
     const char *s, *e, *p;
     long n, limit = 0;
-    VALUE str;
+    VALUE str, lim;
 
-    if (argc == 0) {
+    rb_scan_args(argc, argv, "02", &str, &lim);
+    switch (argc) {
+      case 0:
 	str = rb_rs;
-    }
-    else {
-	VALUE lim, tmp;
+	break;
 
-	rb_scan_args(argc, argv, "11", &str, &lim);
-	if (!NIL_P(lim)) limit = NUM2LONG(lim);
-	else if (!NIL_P(str) && TYPE(str) != T_STRING) {
-	    tmp = rb_check_string_type(str);
+      case 1:
+	if (!NIL_P(str) && !RB_TYPE_P(str, T_STRING)) {
+	    VALUE tmp = rb_check_string_type(str);
 	    if (NIL_P(tmp)) {
 		limit = NUM2LONG(str);
 		if (limit == 0) return rb_str_new(0,0);
@@ -853,9 +980,12 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
 		str = tmp;
 	    }
 	}
-	else if (!NIL_P(str)) {
-	    StringValue(str);
-	}
+	break;
+
+      case 2:
+	if (!NIL_P(str)) StringValue(str);
+	if (!NIL_P(lim)) limit = NUM2LONG(lim);
+	break;
     }
 
     if (ptr->pos >= (n = RSTRING_LEN(ptr->string))) {
@@ -865,7 +995,7 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
     e = s + RSTRING_LEN(ptr->string);
     s += ptr->pos;
     if (limit > 0 && s + limit < e) {
-	e = s + limit;
+	e = rb_enc_right_char_head(s, s + limit, e, rb_enc_get(ptr->string));
     }
     if (NIL_P(str)) {
 	str = strio_substr(ptr, ptr->pos, e - s);
@@ -884,7 +1014,7 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
 		break;
 	    }
 	}
-	str = strio_substr(ptr, s - RSTRING_PTR(ptr->string), e - s); 
+	str = strio_substr(ptr, s - RSTRING_PTR(ptr->string), e - s);
     }
     else if (n == 1) {
 	if ((p = memchr(s, RSTRING_PTR(str)[0], e - s)) != 0) {
@@ -929,7 +1059,7 @@ strio_getline(int argc, VALUE *argv, struct StringIO *ptr)
 static VALUE
 strio_gets(int argc, VALUE *argv, VALUE self)
 {
-    VALUE str = strio_getline(argc, argv, readable(StringIO(self)));
+    VALUE str = strio_getline(argc, argv, readable(self));
 
     rb_lastline_set(str);
     return str;
@@ -946,7 +1076,7 @@ strio_gets(int argc, VALUE *argv, VALUE self)
 static VALUE
 strio_readline(int argc, VALUE *argv, VALUE self)
 {
-    VALUE line = strio_gets(argc, argv, self);
+    VALUE line = rb_funcall2(self, rb_intern("gets"), argc, argv);
     if (NIL_P(line)) rb_eof_error();
     return line;
 }
@@ -956,26 +1086,46 @@ strio_readline(int argc, VALUE *argv, VALUE self)
  *   strio.each(sep=$/) {|line| block }         -> strio
  *   strio.each(limit) {|line| block }          -> strio
  *   strio.each(sep, limit) {|line| block }     -> strio
+ *   strio.each(...)                            -> anEnumerator
+ *
  *   strio.each_line(sep=$/) {|line| block }    -> strio
  *   strio.each_line(limit) {|line| block }     -> strio
  *   strio.each_line(sep,limit) {|line| block } -> strio
+ *   strio.each_line(...)                       -> anEnumerator
  *
  * See IO#each.
  */
 static VALUE
 strio_each(int argc, VALUE *argv, VALUE self)
 {
-    struct StringIO *ptr = StringIO(self);
     VALUE line;
 
+    StringIO(self);
     RETURN_ENUMERATOR(self, argc, argv);
 
-    while (!NIL_P(line = strio_getline(argc, argv, readable(ptr)))) {
+    if (argc > 0 && !NIL_P(argv[argc-1]) && NIL_P(rb_check_string_type(argv[argc-1])) &&
+	NUM2LONG(argv[argc-1]) == 0) {
+	rb_raise(rb_eArgError, "invalid limit: 0 for each_line");
+    }
+
+    while (!NIL_P(line = strio_getline(argc, argv, readable(self)))) {
 	rb_yield(line);
     }
     return self;
 }
 
+/*
+ *  This is a deprecated alias for <code>each_line</code>.
+ */
+static VALUE
+strio_lines(int argc, VALUE *argv, VALUE self)
+{
+    rb_warn("StringIO#lines is deprecated; use #each_line instead");
+    if (!rb_block_given_p())
+	return rb_enumeratorize(self, ID2SYM(rb_intern("each_line")), argc, argv);
+    return strio_each(argc, argv, self);
+}
+
 /*
  * call-seq:
  *   strio.readlines(sep=$/)    ->   array
@@ -987,9 +1137,16 @@ strio_each(int argc, VALUE *argv, VALUE self)
 static VALUE
 strio_readlines(int argc, VALUE *argv, VALUE self)
 {
-    struct StringIO *ptr = StringIO(self);
-    VALUE ary = rb_ary_new(), line;
-    while (!NIL_P(line = strio_getline(argc, argv, readable(ptr)))) {
+    VALUE ary, line;
+
+    StringIO(self);
+    ary = rb_ary_new();
+    if (argc > 0 && !NIL_P(argv[argc-1]) && NIL_P(rb_check_string_type(argv[argc-1])) &&
+	NUM2LONG(argv[argc-1]) == 0) {
+	rb_raise(rb_eArgError, "invalid limit: 0 for readlines");
+    }
+
+    while (!NIL_P(line = strio_getline(argc, argv, readable(self)))) {
 	rb_ary_push(ary, line);
     }
     return ary;
@@ -1008,11 +1165,12 @@ strio_readlines(int argc, VALUE *argv, VALUE self)
 static VALUE
 strio_write(VALUE self, VALUE str)
 {
-    struct StringIO *ptr = writable(StringIO(self));
+    struct StringIO *ptr = writable(self);
     long len, olen;
     rb_encoding *enc, *enc2;
 
-    if (TYPE(str) != T_STRING)
+    RB_GC_GUARD(str);
+    if (!RB_TYPE_P(str, T_STRING))
 	str = rb_obj_as_string(str);
     enc = rb_enc_get(ptr->string);
     enc2 = rb_enc_get(str);
@@ -1027,7 +1185,7 @@ strio_write(VALUE self, VALUE str)
 	ptr->pos = olen;
     }
     if (ptr->pos == olen) {
-	rb_str_cat(ptr->string, RSTRING_PTR(str), len);
+	rb_enc_str_buf_cat(ptr->string, RSTRING_PTR(str), len, enc);
     }
     else {
 	strio_extend(ptr, ptr->pos, len);
@@ -1073,7 +1231,7 @@ strio_write(VALUE self, VALUE str)
 static VALUE
 strio_putc(VALUE self, VALUE ch)
 {
-    struct StringIO *ptr = writable(StringIO(self));
+    struct StringIO *ptr = writable(self);
     int c = NUM2CHR(ch);
     long olen;
 
@@ -1098,25 +1256,28 @@ strio_putc(VALUE self, VALUE ch)
 
 /*
  * call-seq:
- *   strio.read([length [, buffer]])    -> string, buffer, or nil
+ *   strio.read([length [, outbuf]])    -> string, outbuf, or nil
  *
  * See IO#read.
  */
 static VALUE
 strio_read(int argc, VALUE *argv, VALUE self)
 {
-    struct StringIO *ptr = readable(StringIO(self));
+    struct StringIO *ptr = readable(self);
     VALUE str = Qnil;
-    long len, olen;
+    long len;
+    int binary = 0;
 
     switch (argc) {
       case 2:
 	str = argv[1];
-	StringValue(str);
-	rb_str_modify(str);
+	if (!NIL_P(str)) {
+	    StringValue(str);
+	    rb_str_modify(str);
+	}
       case 1:
 	if (!NIL_P(argv[0])) {
-	    len = olen = NUM2LONG(argv[0]);
+	    len = NUM2LONG(argv[0]);
 	    if (len < 0) {
 		rb_raise(rb_eArgError, "negative length %ld given", len);
 	    }
@@ -1124,11 +1285,11 @@ strio_read(int argc, VALUE *argv, VALUE self)
 		if (!NIL_P(str)) rb_str_resize(str, 0);
 		return Qnil;
 	    }
+	    binary = 1;
 	    break;
 	}
 	/* fall through */
       case 0:
-	olen = -1;
 	len = RSTRING_LEN(ptr->string);
 	if (len <= ptr->pos) {
 	    if (NIL_P(str)) {
@@ -1148,26 +1309,26 @@ strio_read(int argc, VALUE *argv, VALUE self)
     }
     if (NIL_P(str)) {
 	str = strio_substr(ptr, ptr->pos, len);
+	if (binary) rb_enc_associate(str, rb_ascii8bit_encoding());
     }
     else {
 	long rest = RSTRING_LEN(ptr->string) - ptr->pos;
 	if (len > rest) len = rest;
 	rb_str_resize(str, len);
 	MEMCPY(RSTRING_PTR(str), RSTRING_PTR(ptr->string) + ptr->pos, char, len);
+	if (binary)
+	    rb_enc_associate(str, rb_ascii8bit_encoding());
+	else
+	    rb_enc_copy(str, ptr->string);
     }
-    if (NIL_P(str)) {
-	str = rb_str_new(0, 0);
-	len = 0;
-    }
-    else {
-	ptr->pos += len = RSTRING_LEN(str);
-    }
+    ptr->pos += RSTRING_LEN(str);
     return str;
 }
 
 /*
  * call-seq:
  *   strio.sysread(integer[, outbuf])    -> string
+ *   strio.readpartial(integer[, outbuf])    -> string
  *
  * Similar to #read, but raises +EOFError+ at end of string instead of
  * returning +nil+, as well as IO#sysread does.
@@ -1175,34 +1336,66 @@ strio_read(int argc, VALUE *argv, VALUE self)
 static VALUE
 strio_sysread(int argc, VALUE *argv, VALUE self)
 {
-    VALUE val = strio_read(argc, argv, self);
+    VALUE val = rb_funcall2(self, rb_intern("read"), argc, argv);
     if (NIL_P(val)) {
 	rb_eof_error();
     }
     return val;
 }
 
-#define strio_syswrite strio_write
-
-/* call-seq: strio.path -> nil */
-#define strio_path strio_nil
-
 /*
  * call-seq:
- *   strio.isatty -> nil
- *   strio.tty? -> nil
+ *   strio.read_nonblock(integer[, outbuf [, opts]])    -> string
  *
+ * Similar to #read, but raises +EOFError+ at end of string unless the
+ * +exception: false+ option is passed in.
  */
+static VALUE
+strio_read_nonblock(int argc, VALUE *argv, VALUE self)
+{
+    VALUE opts = Qnil, val;
+    int no_exception = 0;
+
+    rb_scan_args(argc, argv, "11:", NULL, NULL, &opts);
+
+    if (!NIL_P(opts)) {
+	argc--;
+
+	if (Qfalse == rb_hash_aref(opts, sym_exception))
+	    no_exception = 1;
+    }
+
+    val = strio_read(argc, argv, self);
+    if (NIL_P(val)) {
+	if (no_exception)
+	    return Qnil;
+	else
+	    rb_eof_error();
+    }
+
+    return val;
+}
+
+#define strio_syswrite rb_io_write
+
+static VALUE
+strio_syswrite_nonblock(int argc, VALUE *argv, VALUE self)
+{
+    VALUE str;
+
+    rb_scan_args(argc, argv, "10:", &str, NULL);
+    return strio_syswrite(self, str);
+}
+
 #define strio_isatty strio_false
 
-/* call-seq: strio.pid -> nil */
 #define strio_pid strio_nil
 
-/* call-seq: strio.fileno -> nil */
 #define strio_fileno strio_nil
 
 /*
  * call-seq:
+ *   strio.length -> integer
  *   strio.size   -> integer
  *
  * Returns the size of the buffer string.
@@ -1227,11 +1420,11 @@ strio_size(VALUE self)
 static VALUE
 strio_truncate(VALUE self, VALUE len)
 {
-    VALUE string = writable(StringIO(self))->string;
+    VALUE string = writable(self)->string;
     long l = NUM2LONG(len);
     long plen = RSTRING_LEN(string);
     if (l < 0) {
-	error_inval("negative legnth");
+	error_inval("negative length");
     }
     rb_str_resize(string, l);
     if (plen < l) {
@@ -1270,17 +1463,29 @@ strio_internal_encoding(VALUE self)
 
 /*
  *  call-seq:
- *     strio.set_encoding(ext_enc)                => strio
+ *     strio.set_encoding(ext_enc, [int_enc[, opt]])  => strio
  *
- *  Tagged with the encoding specified.
+ *  Specify the encoding of the StringIO as <i>ext_enc</i>.
+ *  Use the default external encoding if <i>ext_enc</i> is nil.
+ *  2nd argument <i>int_enc</i> and optional hash <i>opt</i> argument
+ *  are ignored; they are for API compatibility to IO.
  */
 
 static VALUE
-strio_set_encoding(VALUE self, VALUE ext_enc)
+strio_set_encoding(int argc, VALUE *argv, VALUE self)
 {
     rb_encoding* enc;
     VALUE str = StringIO(self)->string;
-    enc = rb_to_encoding(ext_enc);
+    VALUE ext_enc, int_enc, opt;
+
+    argc = rb_scan_args(argc, argv, "11:", &ext_enc, &int_enc, &opt);
+
+    if (NIL_P(ext_enc)) {
+	enc = rb_default_external_encoding();
+    }
+    else {
+	enc = rb_to_encoding(ext_enc);
+    }
     rb_enc_associate(str, enc);
     return self;
 }
@@ -1305,6 +1510,8 @@ Init_stringio()
     rb_define_method(StringIO, "lineno", strio_get_lineno, 0);
     rb_define_method(StringIO, "lineno=", strio_set_lineno, 1);
 
+
+    /* call-seq: strio.binmode -> true */
     rb_define_method(StringIO, "binmode", strio_binmode, 0);
     rb_define_method(StringIO, "close", strio_close, 0);
     rb_define_method(StringIO, "close_read", strio_close_read, 0);
@@ -1314,49 +1521,54 @@ Init_stringio()
     rb_define_method(StringIO, "closed_write?", strio_closed_write, 0);
     rb_define_method(StringIO, "eof", strio_eof, 0);
     rb_define_method(StringIO, "eof?", strio_eof, 0);
+    /* call-seq: strio.fcntl */
     rb_define_method(StringIO, "fcntl", strio_fcntl, -1);
+    /* call-seq: strio.flush -> strio */
     rb_define_method(StringIO, "flush", strio_flush, 0);
+    /* call-seq: strio.fsync -> 0 */
     rb_define_method(StringIO, "fsync", strio_fsync, 0);
     rb_define_method(StringIO, "pos", strio_get_pos, 0);
     rb_define_method(StringIO, "pos=", strio_set_pos, 1);
     rb_define_method(StringIO, "rewind", strio_rewind, 0);
     rb_define_method(StringIO, "seek", strio_seek, -1);
     rb_define_method(StringIO, "sync", strio_get_sync, 0);
+    /* call-seq: strio.sync = boolean -> boolean */
     rb_define_method(StringIO, "sync=", strio_set_sync, 1);
     rb_define_method(StringIO, "tell", strio_tell, 0);
-    rb_define_method(StringIO, "path", strio_path, 0);
 
     rb_define_method(StringIO, "each", strio_each, -1);
     rb_define_method(StringIO, "each_line", strio_each, -1);
-    rb_define_method(StringIO, "lines", strio_each, -1);
+    rb_define_method(StringIO, "lines", strio_lines, -1);
     rb_define_method(StringIO, "each_byte", strio_each_byte, 0);
-    rb_define_method(StringIO, "bytes", strio_each_byte, 0);
+    rb_define_method(StringIO, "bytes", strio_bytes, 0);
     rb_define_method(StringIO, "each_char", strio_each_char, 0);
-    rb_define_method(StringIO, "chars", strio_each_char, 0);
+    rb_define_method(StringIO, "chars", strio_chars, 0);
+    rb_define_method(StringIO, "each_codepoint", strio_each_codepoint, 0);
+    rb_define_method(StringIO, "codepoints", strio_codepoints, 0);
     rb_define_method(StringIO, "getc", strio_getc, 0);
     rb_define_method(StringIO, "ungetc", strio_ungetc, 1);
     rb_define_method(StringIO, "ungetbyte", strio_ungetbyte, 1);
-    rb_define_method(StringIO, "readchar", strio_readchar, 0);
     rb_define_method(StringIO, "getbyte", strio_getbyte, 0);
-    rb_define_method(StringIO, "readbyte", strio_readbyte, 0);
     rb_define_method(StringIO, "gets", strio_gets, -1);
-    rb_define_method(StringIO, "readline", strio_readline, -1);
     rb_define_method(StringIO, "readlines", strio_readlines, -1);
     rb_define_method(StringIO, "read", strio_read, -1);
-    rb_define_method(StringIO, "sysread", strio_sysread, -1);
-    rb_define_method(StringIO, "readpartial", strio_sysread, -1);
 
     rb_define_method(StringIO, "write", strio_write, 1);
-    rb_define_method(StringIO, "<<", strio_addstr, 1);
-    rb_define_method(StringIO, "print", strio_print, -1);
-    rb_define_method(StringIO, "printf", strio_printf, -1);
     rb_define_method(StringIO, "putc", strio_putc, 1);
-    rb_define_method(StringIO, "puts", strio_puts, -1);
-    rb_define_method(StringIO, "syswrite", strio_syswrite, 1);
 
+    /*
+     * call-seq:
+     *   strio.isatty -> nil
+     *   strio.tty? -> nil
+     *
+     */
     rb_define_method(StringIO, "isatty", strio_isatty, 0);
     rb_define_method(StringIO, "tty?", strio_isatty, 0);
+
+    /* call-seq: strio.pid -> nil */
     rb_define_method(StringIO, "pid", strio_pid, 0);
+
+    /* call-seq: strio.fileno -> nil */
     rb_define_method(StringIO, "fileno", strio_fileno, 0);
     rb_define_method(StringIO, "size", strio_size, 0);
     rb_define_method(StringIO, "length", strio_size, 0);
@@ -1364,5 +1576,28 @@ Init_stringio()
 
     rb_define_method(StringIO, "external_encoding", strio_external_encoding, 0);
     rb_define_method(StringIO, "internal_encoding", strio_internal_encoding, 0);
-    rb_define_method(StringIO, "set_encoding", strio_set_encoding, 1);
+    rb_define_method(StringIO, "set_encoding", strio_set_encoding, -1);
+
+    {
+	VALUE mReadable = rb_define_module_under(rb_cIO, "generic_readable");
+	rb_define_method(mReadable, "readchar", strio_readchar, 0);
+	rb_define_method(mReadable, "readbyte", strio_readbyte, 0);
+	rb_define_method(mReadable, "readline", strio_readline, -1);
+	rb_define_method(mReadable, "sysread", strio_sysread, -1);
+	rb_define_method(mReadable, "readpartial", strio_sysread, -1);
+	rb_define_method(mReadable, "read_nonblock", strio_read_nonblock, -1);
+	rb_include_module(StringIO, mReadable);
+    }
+    {
+	VALUE mWritable = rb_define_module_under(rb_cIO, "generic_writable");
+	rb_define_method(mWritable, "<<", strio_addstr, 1);
+	rb_define_method(mWritable, "print", strio_print, -1);
+	rb_define_method(mWritable, "printf", strio_printf, -1);
+	rb_define_method(mWritable, "puts", strio_puts, -1);
+	rb_define_method(mWritable, "syswrite", strio_syswrite, 1);
+	rb_define_method(mWritable, "write_nonblock", strio_syswrite_nonblock, -1);
+	rb_include_module(StringIO, mWritable);
+    }
+
+    sym_exception = ID2SYM(rb_intern("exception"));
 }
