@@ -1,5 +1,5 @@
 /*
-    $Id: strscan.c 34053 2011-12-15 05:48:39Z nobu $
+    $Id: strscan.c 40927 2013-05-25 11:37:35Z naruse $
 
     Copyright (c) 1999-2006 Minero Aoki
 
@@ -21,6 +21,7 @@
 
 static VALUE StringScanner;
 static VALUE ScanError;
+static ID id_byteslice;
 
 struct strscanner
 {
@@ -37,6 +38,9 @@ struct strscanner
 
     /* the regexp register; legal only when MATCHED_P(s) */
     struct re_registers regs;
+
+    /* regexp used for last scan */
+    VALUE regex;
 };
 
 #define MATCHED_P(s)          ((s)->flags & FLAG_MATCHED)
@@ -371,7 +375,7 @@ strscan_concat(VALUE self, VALUE str)
  * value is zero.  In the 'terminated' position (i.e. the string is exhausted),
  * this value is the bytesize of the string.
  *
- * In short, it's a 0-based index into the string.
+ * In short, it's a 0-based index into bytes of the string.
  *
  *   s = StringScanner.new('test string')
  *   s.pos               # -> 0
@@ -389,6 +393,32 @@ strscan_get_pos(VALUE self)
     return INT2FIX(p->curr);
 }
 
+/*
+ * Returns the character position of the scan pointer.  In the 'reset' position, this
+ * value is zero.  In the 'terminated' position (i.e. the string is exhausted),
+ * this value is the size of the string.
+ *
+ * In short, it's a 0-based index into the string.
+ *
+ *   s = StringScanner.new("abcädeföghi")
+ *   s.charpos           # -> 0
+ *   s.scan_until(/ä/)   # -> "abcä"
+ *   s.pos               # -> 5
+ *   s.charpos           # -> 4
+ */
+static VALUE
+strscan_get_charpos(VALUE self)
+{
+    struct strscanner *p;
+    VALUE substr;
+
+    GET_SCANNER(self, p);
+
+    substr = rb_funcall(p->str, id_byteslice, 2, INT2FIX(0), INT2NUM(p->curr));
+
+    return rb_str_length(substr);
+}
+
 /*
  * call-seq: pos=(n)
  *
@@ -429,6 +459,8 @@ strscan_do_scan(VALUE self, VALUE regex, int succptr, int getstr, int headonly)
     if (S_RESTLEN(p) < 0) {
         return Qnil;
     }
+
+    p->regex = regex;
     re = rb_reg_prepare_re(regex, p->str);
     tmpreg = re != RREGEXP(regex)->ptr;
     if (!tmpreg) RREGEXP(regex)->usecnt++;
@@ -943,6 +975,25 @@ strscan_matched_size(VALUE self)
     return INT2NUM(p->regs.end[0] - p->regs.beg[0]);
 }
 
+static int
+name_to_backref_number(struct re_registers *regs, VALUE regexp, const char* name, const char* name_end)
+{
+    int num;
+
+    num = onig_name_to_backref_number(RREGEXP(regexp)->ptr,
+	(const unsigned char* )name, (const unsigned char* )name_end, regs);
+    if (num >= 1) {
+	return num;
+    }
+    else {
+	VALUE s = rb_str_new(name, (long )(name_end - name));
+	rb_raise(rb_eIndexError, "undefined group name reference: %s",
+				 StringValuePtr(s));
+    }
+
+    UNREACHABLE;
+}
+
 /*
  * call-seq: [](n)
  *
@@ -956,17 +1007,43 @@ strscan_matched_size(VALUE self)
  *   s[3]                               # -> "12"
  *   s.post_match                       # -> "1975 14:39"
  *   s.pre_match                        # -> ""
+ *
+ *   s.reset
+ *   s.scan(/(?<wday>\w+) (?<month>\w+) (?<day>\d+) /)       # -> "Fri Dec 12 "
+ *   s[0]                               # -> "Fri Dec 12 "
+ *   s[1]                               # -> "Fri"
+ *   s[2]                               # -> "Dec"
+ *   s[3]                               # -> "12"
+ *   s[:wday]                           # -> "Fri"
+ *   s[:month]                          # -> "Dec"
+ *   s[:day]                            # -> "12"
+ *   s.post_match                       # -> "1975 14:39"
+ *   s.pre_match                        # -> ""
  */
 static VALUE
 strscan_aref(VALUE self, VALUE idx)
 {
+    const char *name;
     struct strscanner *p;
     long i;
 
     GET_SCANNER(self, p);
     if (! MATCHED_P(p))        return Qnil;
 
-    i = NUM2LONG(idx);
+    switch (TYPE(idx)) {
+        case T_SYMBOL:
+            name = rb_id2name(SYM2ID(idx));
+            goto name_to_backref;
+            break;
+        case T_STRING:
+            name = StringValuePtr(idx);
+        name_to_backref:
+	    i = name_to_backref_number(&(p->regs), p->regex, name, name + strlen(name));
+            break;
+        default:
+            i = NUM2LONG(idx);
+    }
+
     if (i < 0)
         i += p->regs.num_regs;
     if (i < 0)                 return Qnil;
@@ -1262,6 +1339,8 @@ Init_strscan()
     ID id_scanerr = rb_intern("ScanError");
     VALUE tmp;
 
+    id_byteslice = rb_intern("byteslice");
+
     StringScanner = rb_define_class("StringScanner", rb_cObject);
     ScanError = rb_define_class_under(StringScanner, "Error", rb_eStandardError);
     if (!rb_const_defined(rb_cObject, id_scanerr)) {
@@ -1270,7 +1349,7 @@ Init_strscan()
     tmp = rb_str_new2(STRSCAN_VERSION);
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Version"), tmp);
-    tmp = rb_str_new2("$Id: strscan.c 34053 2011-12-15 05:48:39Z nobu $");
+    tmp = rb_str_new2("$Id: strscan.c 40927 2013-05-25 11:37:35Z naruse $");
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Id"), tmp);
 
@@ -1287,6 +1366,7 @@ Init_strscan()
     rb_define_method(StringScanner, "<<",          strscan_concat,      1);
     rb_define_method(StringScanner, "pos",         strscan_get_pos,     0);
     rb_define_method(StringScanner, "pos=",        strscan_set_pos,     1);
+    rb_define_method(StringScanner, "charpos",     strscan_get_charpos, 0);
     rb_define_method(StringScanner, "pointer",     strscan_get_pos,     0);
     rb_define_method(StringScanner, "pointer=",    strscan_set_pos,     1);
 
