@@ -2,7 +2,7 @@
 
   iseq.c -
 
-  $Author: nobu $
+  $Author: ko1 $
   created at: 2006-07-11(Tue) 09:00:03 +0900
 
   Copyright (C) 2006 Koichi Sasada
@@ -18,13 +18,11 @@
 #include "vm_core.h"
 #include "iseq.h"
 
-#define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
-
 #include "insns.inc"
 #include "insns_info.inc"
 
 #define ISEQ_MAJOR_VERSION 2
-#define ISEQ_MINOR_VERSION 0
+#define ISEQ_MINOR_VERSION 1
 
 VALUE rb_cISeq;
 
@@ -83,7 +81,7 @@ iseq_free(void *ptr)
 	    RUBY_FREE_UNLESS_NULL(iseq->iseq);
 	    RUBY_FREE_UNLESS_NULL(iseq->line_info_table);
 	    RUBY_FREE_UNLESS_NULL(iseq->local_table);
-	    RUBY_FREE_UNLESS_NULL(iseq->ic_entries);
+	    RUBY_FREE_UNLESS_NULL(iseq->is_entries);
 	    RUBY_FREE_UNLESS_NULL(iseq->callinfo_entries);
 	    RUBY_FREE_UNLESS_NULL(iseq->catch_table);
 	    RUBY_FREE_UNLESS_NULL(iseq->arg_opt_table);
@@ -114,10 +112,6 @@ iseq_mark(void *ptr)
 	RUBY_MARK_UNLESS_NULL((VALUE)iseq->cref_stack);
 	RUBY_MARK_UNLESS_NULL(iseq->klass);
 	RUBY_MARK_UNLESS_NULL(iseq->coverage);
-#if 0
-	RUBY_MARK_UNLESS_NULL((VALUE)iseq->node);
-	RUBY_MARK_UNLESS_NULL(iseq->cached_special_block);
-#endif
 	RUBY_MARK_UNLESS_NULL(iseq->orig);
 
 	if (iseq->compile_data != 0) {
@@ -148,7 +142,7 @@ iseq_memsize(const void *ptr)
 	    size += iseq->local_table_size * sizeof(ID);
 	    size += iseq->catch_table_size * sizeof(struct iseq_catch_table_entry);
 	    size += iseq->arg_opts * sizeof(VALUE);
-	    size += iseq->ic_size * sizeof(struct iseq_inline_cache_entry);
+	    size += iseq->is_size * sizeof(union iseq_inline_storage_entry);
 	    size += iseq->callinfo_size * sizeof(rb_call_info_t);
 
 	    if (iseq->compile_data) {
@@ -173,7 +167,10 @@ static const rb_data_type_t iseq_data_type = {
 	iseq_mark,
 	iseq_free,
 	iseq_memsize,
-    },
+    },              /* functions */
+    0,              /* parent */
+    0,              /* data */
+    FL_WB_PROTECTED /* flags */
 };
 
 static VALUE
@@ -187,13 +184,21 @@ static rb_iseq_location_t *
 iseq_location_setup(rb_iseq_t *iseq, VALUE path, VALUE absolute_path, VALUE name, size_t first_lineno)
 {
     rb_iseq_location_t *loc = &iseq->location;
-    loc->path = path;
-    loc->absolute_path = absolute_path;
-    loc->label = loc->base_label = name;
+    OBJ_WRITE(iseq->self, &loc->path, path);
+    if (RTEST(absolute_path) && rb_str_cmp(path, absolute_path) == 0) {
+	OBJ_WRITE(iseq->self, &loc->absolute_path, path);
+    }
+    else {
+	OBJ_WRITE(iseq->self, &loc->absolute_path, absolute_path);
+    }
+    OBJ_WRITE(iseq->self, &loc->label, name);
+    OBJ_WRITE(iseq->self, &loc->base_label, name);
     loc->first_lineno = first_lineno;
     return loc;
 }
 
+#define ISEQ_SET_CREF(iseq, cref) OBJ_WRITE((iseq)->self, &(iseq)->cref_stack, (cref))
+
 static void
 set_relation(rb_iseq_t *iseq, const VALUE parent)
 {
@@ -204,7 +209,7 @@ set_relation(rb_iseq_t *iseq, const VALUE parent)
     /* set class nest stack */
     if (type == ISEQ_TYPE_TOP) {
 	/* toplevel is private */
-	iseq->cref_stack = NEW_CREF(rb_cObject);
+	OBJ_WRITE(iseq->self, &iseq->cref_stack, NEW_CREF(rb_cObject));
 	iseq->cref_stack->nd_refinements = Qnil;
 	iseq->cref_stack->nd_visi = NOEX_PRIVATE;
 	if (th->top_wrapper) {
@@ -212,18 +217,18 @@ set_relation(rb_iseq_t *iseq, const VALUE parent)
 	    cref->nd_refinements = Qnil;
 	    cref->nd_visi = NOEX_PRIVATE;
 	    cref->nd_next = iseq->cref_stack;
-	    iseq->cref_stack = cref;
+	    ISEQ_SET_CREF(iseq, cref);
 	}
 	iseq->local_iseq = iseq;
     }
     else if (type == ISEQ_TYPE_METHOD || type == ISEQ_TYPE_CLASS) {
-	iseq->cref_stack = NEW_CREF(0); /* place holder */
+	ISEQ_SET_CREF(iseq, NEW_CREF(0)); /* place holder */
 	iseq->cref_stack->nd_refinements = Qnil;
 	iseq->local_iseq = iseq;
     }
     else if (RTEST(parent)) {
 	GetISeqPtr(parent, piseq);
-	iseq->cref_stack = piseq->cref_stack;
+	ISEQ_SET_CREF(iseq, piseq->cref_stack);
 	iseq->local_iseq = piseq->local_iseq;
     }
 
@@ -231,6 +236,20 @@ set_relation(rb_iseq_t *iseq, const VALUE parent)
 	GetISeqPtr(parent, piseq);
 	iseq->parent_iseq = piseq;
     }
+
+    if (type == ISEQ_TYPE_MAIN) {
+	iseq->local_iseq = iseq;
+    }
+}
+
+void
+rb_iseq_add_mark_object(rb_iseq_t *iseq, VALUE obj)
+{
+    if (!RTEST(iseq->mark_ary)) {
+	OBJ_WRITE(iseq->self, &iseq->mark_ary, rb_ary_tmp_new(3));
+	RBASIC_CLEAR_CLASS(iseq->mark_ary);
+    }
+    rb_ary_push(iseq->mark_ary, obj);
 }
 
 static VALUE
@@ -243,7 +262,7 @@ prepare_iseq_build(rb_iseq_t *iseq,
     iseq->arg_rest = -1;
     iseq->arg_block = -1;
     iseq->arg_keyword = -1;
-    iseq->klass = 0;
+    OBJ_WRITE(iseq->self, &iseq->klass, 0);
     set_relation(iseq, parent);
 
     OBJ_FREEZE(name);
@@ -251,14 +270,11 @@ prepare_iseq_build(rb_iseq_t *iseq,
 
     iseq_location_setup(iseq, path, absolute_path, name, first_lineno);
     if (iseq != iseq->local_iseq) {
-	iseq->location.base_label = iseq->local_iseq->location.label;
+	OBJ_WRITE(iseq->self, &iseq->location.base_label, iseq->local_iseq->location.label);
     }
 
     iseq->defined_method_id = 0;
-    iseq->mark_ary = rb_ary_tmp_new(3);
-    OBJ_UNTRUST(iseq->mark_ary);
-    RBASIC(iseq->mark_ary)->klass = 0;
-
+    OBJ_WRITE(iseq->self, &iseq->mark_ary, 0);
 
     /*
      * iseq->special_block_builder = GC_GUARDED_PTR_REF(block_opt);
@@ -268,15 +284,15 @@ prepare_iseq_build(rb_iseq_t *iseq,
 
     iseq->compile_data = ALLOC(struct iseq_compile_data);
     MEMZERO(iseq->compile_data, struct iseq_compile_data, 1);
-    iseq->compile_data->err_info = Qnil;
-    iseq->compile_data->mark_ary = rb_ary_tmp_new(3);
+    OBJ_WRITE(iseq->self, &iseq->compile_data->err_info, Qnil);
+    OBJ_WRITE(iseq->self, &iseq->compile_data->mark_ary, rb_ary_tmp_new(3));
 
     iseq->compile_data->storage_head = iseq->compile_data->storage_current =
       (struct iseq_compile_data_storage *)
 	ALLOC_N(char, INITIAL_ISEQ_COMPILE_DATA_STORAGE_BUFF_SIZE +
 		sizeof(struct iseq_compile_data_storage));
 
-    iseq->compile_data->catch_table_ary = rb_ary_new();
+    OBJ_WRITE(iseq->self, &iseq->compile_data->catch_table_ary, rb_ary_new());
     iseq->compile_data->storage_head->pos = 0;
     iseq->compile_data->storage_head->next = 0;
     iseq->compile_data->storage_head->size =
@@ -286,12 +302,12 @@ prepare_iseq_build(rb_iseq_t *iseq,
     iseq->compile_data->option = option;
     iseq->compile_data->last_coverable_line = -1;
 
-    iseq->coverage = Qfalse;
+    OBJ_WRITE(iseq->self, &iseq->coverage, Qfalse);
     if (!GET_THREAD()->parse_in_eval) {
 	VALUE coverages = rb_get_coverages();
 	if (RTEST(coverages)) {
-	    iseq->coverage = rb_hash_lookup(coverages, path);
-	    if (NIL_P(iseq->coverage)) iseq->coverage = Qfalse;
+	    OBJ_WRITE(iseq->self, &iseq->coverage, rb_hash_lookup(coverages, path));
+	    if (NIL_P(iseq->coverage)) OBJ_WRITE(iseq->self, &iseq->coverage, Qfalse);
 	}
     }
 
@@ -382,6 +398,7 @@ make_compile_option_value(rb_compile_option_t *option)
 	SET_COMPILE_OPTION(option, opt, operands_unification);
 	SET_COMPILE_OPTION(option, opt, instructions_unification);
 	SET_COMPILE_OPTION(option, opt, stack_caching);
+	SET_COMPILE_OPTION(option, opt, trace_instruction);
 	SET_COMPILE_OPTION_NUM(option, opt, debug_level);
     }
 #undef SET_COMPILE_OPTION
@@ -501,6 +518,7 @@ iseq_load(VALUE self, VALUE data, VALUE parent, VALUE opt)
 
     GetISeqPtr(iseqval, iseq);
     iseq->self = iseqval;
+    iseq->local_iseq = iseq;
 
     if (type_map == 0) {
 	type_map = st_init_numtable();
@@ -556,18 +574,6 @@ rb_iseq_load(VALUE data, VALUE parent, VALUE opt)
     return iseq_load(rb_cISeq, data, parent, opt);
 }
 
-static NODE *
-parse_string(VALUE str, const char *file, int line)
-{
-    VALUE parser = rb_parser_new();
-    NODE *node = rb_parser_compile_string(parser, file, str, line);
-
-    if (!node) {
-	rb_exc_raise(GET_THREAD()->errinfo);	/* TODO: check err */
-    }
-    return node;
-}
-
 VALUE
 rb_iseq_compile_with_option(VALUE src, VALUE file, VALUE absolute_path, VALUE line, rb_block_t *base_block, VALUE opt)
 {
@@ -580,17 +586,25 @@ rb_iseq_compile_with_option(VALUE src, VALUE file, VALUE absolute_path, VALUE li
 
     TH_PUSH_TAG(th);
     if ((state = EXEC_TAG()) == 0) {
+	VALUE parser;
 	int ln = NUM2INT(line);
-	const char *fn = StringValueCStr(file);
 	NODE *node;
 	rb_compile_option_t option;
 
+	StringValueCStr(file);
 	make_compile_option(&option, opt);
 
+	parser = rb_parser_new();
+
 	if (RB_TYPE_P((src), T_FILE))
-	    node = rb_compile_file(fn, src, ln);
-	else
-	    node = parse_string(StringValue(src), fn, ln);
+	    node = rb_parser_compile_file_path(parser, file, src, ln);
+	else {
+	    node = rb_parser_compile_string_path(parser, file, src, ln);
+
+	    if (!node) {
+		rb_exc_raise(GET_THREAD()->errinfo);	/* TODO: check err */
+	    }
+	}
 
 	if (base_block && base_block->iseq) {
 	    iseqval = rb_iseq_new_with_opt(node, base_block->iseq->location.label,
@@ -638,8 +652,9 @@ rb_iseq_compile_on_base(VALUE src, VALUE file, VALUE line, rb_block_t *base_bloc
  *  metadata attached to the returned +iseq+.
  *
  *  +options+, which can be +true+, +false+ or a +Hash+, is used to
- *  modify the default behavior of the Ruby iseq compiler.  For details
- *  regarding valid compile options see InstructionSequence.compile_option=.
+ *  modify the default behavior of the Ruby iseq compiler.
+ *
+ *  For details regarding valid compile options see ::compile_option=.
  *
  *     RubyVM::InstructionSequence.compile("a = 1 + 2")
  *     #=> <RubyVM::InstructionSequence:<compiled>@<compiled>>
@@ -668,8 +683,9 @@ iseq_s_compile(int argc, VALUE *argv, VALUE self)
  *  InstructionSequence with source location metadata set.
  *
  *  Optionally takes +options+, which can be +true+, +false+ or a +Hash+, to
- *  modify the default behavior of the Ruby iseq compiler (for details see
- *  InstructionSequence.compile_option=).
+ *  modify the default behavior of the Ruby iseq compiler.
+ *
+ *  For details regarding valid compile options see ::compile_option=.
  *
  *      # /tmp/hello.rb
  *      puts "Hello, world!"
@@ -708,10 +724,14 @@ iseq_s_compile_file(int argc, VALUE *argv, VALUE self)
  *     InstructionSequence.compile_option = options
  *
  *  Sets the default values for various optimizations in the Ruby iseq
- *  compiler.  Possible values for +options+ include +true+, which enables all
- *  options, +false+ which disables all options, a +Hash+ of options that you
- *  want to change (options not present in the hash will be left unchanged),
- *  and +nil+ which leaves all options unchanged.
+ *  compiler.
+ *
+ *  Possible values for +options+ include +true+, which enables all options,
+ *  +false+ which disables all options, and +nil+ which leaves all options
+ *  unchanged.
+ *
+ *  You can also pass a +Hash+ of +options+ that you want to change, any
+ *  options not present in the hash will be left unchanged.
  *
  *  Possible option names (which are keys in +options+) which can be set to
  *  +true+ or +false+ include:
@@ -729,8 +749,7 @@ iseq_s_compile_file(int argc, VALUE *argv, VALUE self)
  *
  *  These default options can be overwritten for a single run of the iseq
  *  compiler by passing any of the above values as the +options+ parameter to
- *  InstructionSequence.new, InstructionSequence.compile and
- *  InstructionSequence.compile_file.
+ *  ::new, ::compile and ::compile_file.
  */
 static VALUE
 iseq_s_compile_option_set(VALUE self, VALUE opt)
@@ -747,6 +766,7 @@ iseq_s_compile_option_set(VALUE self, VALUE opt)
  *     InstructionSequence.compile_option -> options
  *
  *  Returns a hash of default options used by the Ruby iseq compiler.
+ *
  *  For details, see InstructionSequence.compile_option=.
  */
 static VALUE
@@ -782,7 +802,8 @@ iseq_eval(VALUE self)
 }
 
 /*
- *  :nodoc:
+ *  Returns a human-readable string representation of this instruction
+ *  sequence, including the #label and #path.
  */
 static VALUE
 iseq_inspect(VALUE self)
@@ -798,6 +819,138 @@ iseq_inspect(VALUE self)
 		      RSTRING_PTR(iseq->location.label), RSTRING_PTR(iseq->location.path));
 }
 
+/*
+ *  Returns the path of this instruction sequence.
+ *
+ *  <code><compiled></code> if the iseq was evaluated from a string.
+ *
+ *  For example, using irb:
+ *
+ *	iseq = RubyVM::InstructionSequence.compile('num = 1 + 2')
+ *	#=> <RubyVM::InstructionSequence:<compiled>@<compiled>>
+ *	iseq.path
+ *	#=> "<compiled>"
+ *
+ *  Using ::compile_file:
+ *
+ *	# /tmp/method.rb
+ *	def hello
+ *	  puts "hello, world"
+ *	end
+ *
+ *	# in irb
+ *	> iseq = RubyVM::InstructionSequence.compile_file('/tmp/method.rb')
+ *	> iseq.path #=> /tmp/method.rb
+ */
+static VALUE
+iseq_path(VALUE self)
+{
+    rb_iseq_t *iseq;
+    GetISeqPtr(self, iseq);
+    return iseq->location.path;
+}
+
+/*
+ *  Returns the absolute path of this instruction sequence.
+ *
+ *  +nil+ if the iseq was evaluated from a string.
+ *
+ *  For example, using ::compile_file:
+ *
+ *	# /tmp/method.rb
+ *	def hello
+ *	  puts "hello, world"
+ *	end
+ *
+ *	# in irb
+ *	> iseq = RubyVM::InstructionSequence.compile_file('/tmp/method.rb')
+ *	> iseq.absolute_path #=> /tmp/method.rb
+ */
+static VALUE
+iseq_absolute_path(VALUE self)
+{
+    rb_iseq_t *iseq;
+    GetISeqPtr(self, iseq);
+    return iseq->location.absolute_path;
+}
+
+/*  Returns the label of this instruction sequence.
+ *
+ *  <code><main></code> if it's at the top level, <code><compiled></code> if it
+ *  was evaluated from a string.
+ *
+ *  For example, using irb:
+ *
+ *	iseq = RubyVM::InstructionSequence.compile('num = 1 + 2')
+ *	#=> <RubyVM::InstructionSequence:<compiled>@<compiled>>
+ *	iseq.label
+ *	#=> "<compiled>"
+ *
+ *  Using ::compile_file:
+ *
+ *	# /tmp/method.rb
+ *	def hello
+ *	  puts "hello, world"
+ *	end
+ *
+ *	# in irb
+ *	> iseq = RubyVM::InstructionSequence.compile_file('/tmp/method.rb')
+ *	> iseq.label #=> <main>
+ */
+static VALUE
+iseq_label(VALUE self)
+{
+    rb_iseq_t *iseq;
+    GetISeqPtr(self, iseq);
+    return iseq->location.label;
+}
+
+/*  Returns the base label of this instruction sequence.
+ *
+ *  For example, using irb:
+ *
+ *	iseq = RubyVM::InstructionSequence.compile('num = 1 + 2')
+ *	#=> <RubyVM::InstructionSequence:<compiled>@<compiled>>
+ *	iseq.base_label
+ *	#=> "<compiled>"
+ *
+ *  Using ::compile_file:
+ *
+ *	# /tmp/method.rb
+ *	def hello
+ *	  puts "hello, world"
+ *	end
+ *
+ *	# in irb
+ *	> iseq = RubyVM::InstructionSequence.compile_file('/tmp/method.rb')
+ *	> iseq.base_label #=> <main>
+ */
+static VALUE
+iseq_base_label(VALUE self)
+{
+    rb_iseq_t *iseq;
+    GetISeqPtr(self, iseq);
+    return iseq->location.base_label;
+}
+
+/*  Returns the number of the first source line where the instruction sequence
+ *  was loaded from.
+ *
+ *  For example, using irb:
+ *
+ *	iseq = RubyVM::InstructionSequence.compile('num = 1 + 2')
+ *	#=> <RubyVM::InstructionSequence:<compiled>@<compiled>>
+ *	iseq.first_lineno
+ *	#=> 1
+ */
+static VALUE
+iseq_first_lineno(VALUE self)
+{
+    rb_iseq_t *iseq;
+    GetISeqPtr(self, iseq);
+    return iseq->location.first_lineno;
+}
+
 static
 VALUE iseq_data_to_ary(rb_iseq_t *iseq);
 
@@ -805,12 +958,12 @@ VALUE iseq_data_to_ary(rb_iseq_t *iseq);
  *  call-seq:
  *     iseq.to_a -> ary
  *
- *  Returns an Array of size 14 representing the instruction sequence with
- *  the following data:
+ *  Returns an Array with 14 elements representing the instruction sequence
+ *  with the following data:
  *
  *  [magic]
- *    A string identifying the data format. Always
- *    "YARVInstructionSequence/SimpleDataFormat".
+ *    A string identifying the data format. <b>Always
+ *    +YARVInstructionSequence/SimpleDataFormat+.</b>
  *
  *  [major_version]
  *    The major version of the instruction sequence.
@@ -819,45 +972,57 @@ VALUE iseq_data_to_ary(rb_iseq_t *iseq);
  *    The minor version of the instruction sequence.
  *
  *  [format_type]
- *    A number identifying the data format. Always 1.
+ *    A number identifying the data format. <b>Always 1</b>.
  *
  *  [misc]
- *    A hash containing +:arg_size+, the total number of arguments taken by
- *    the method or the block (0 if _iseq_ doesn't represent a method or
- *    block), +:local_size+, the number of local variables + 1, and
- *    +:stack_max+, used in calculating the stack depth at which a
- *    +SystemStackError+ is thrown.
+ *    A hash containing:
  *
- *  [name]
+ *    [+:arg_size+]
+ *	the total number of arguments taken by the method or the block (0 if
+ *	_iseq_ doesn't represent a method or block)
+ *    [+:local_size+]
+ *	the number of local variables + 1
+ *    [+:stack_max+]
+ *	used in calculating the stack depth at which a SystemStackError is
+ *	thrown.
+ *
+ *  [#label]
  *    The name of the context (block, method, class, module, etc.) that this
- *    instruction sequence belongs to. <code><main></code> if it's at the top
- *    level, <code><compiled></code> if it was evaluated from a string.
+ *    instruction sequence belongs to.
  *
- *  [path]
+ *    <code><main></code> if it's at the top level, <code><compiled></code> if
+ *    it was evaluated from a string.
+ *
+ *  [#path]
  *    The relative path to the Ruby file where the instruction sequence was
- *    loaded from. <code><compiled></code> if the iseq was evaluated from a
- *    string.
+ *    loaded from.
  *
- *  [absolute_path]
+ *    <code><compiled></code> if the iseq was evaluated from a string.
+ *
+ *  [#absolute_path]
  *    The absolute path to the Ruby file where the instruction sequence was
- *    loaded from. +nil+ if the iseq was evaluated from a string.
+ *    loaded from.
  *
- *  [start_lineno]
+ *    +nil+ if the iseq was evaluated from a string.
+ *
+ *  [#first_lineno]
  *    The number of the first source line where the instruction sequence was
  *    loaded from.
  *
  *  [type]
- *    The type of the instruction sequence. Valid values are +:top+,
- *    +:method+, +:block+, +:class+, +:rescue+, +:ensure+, +:eval+, +:main+,
- *    and +:defined_guard+.
+ *    The type of the instruction sequence.
+ *
+ *    Valid values are +:top+, +:method+, +:block+, +:class+, +:rescue+,
+ *    +:ensure+, +:eval+, +:main+, and +:defined_guard+.
  *
  *  [locals]
  *    An array containing the names of all arguments and local variables as
  *    symbols.
  *
  *  [args]
- *    The arity if the method or block only has required arguments.  Otherwise
- *    an array of:
+ *    The arity if the method or block only has required arguments.
+ *
+ *    Otherwise an array of:
  *
  *      [required_argc, [optional_arg_labels, ...],
  *       splat_index, post_splat_argc, post_splat_index,
@@ -963,9 +1128,9 @@ id_to_name(ID id, VALUE default_value)
 }
 
 VALUE
-insn_operand_intern(rb_iseq_t *iseq,
-		    VALUE insn, int op_no, VALUE op,
-		    int len, size_t pos, VALUE *pnop, VALUE child)
+rb_insn_operand_intern(rb_iseq_t *iseq,
+		       VALUE insn, int op_no, VALUE op,
+		       int len, size_t pos, VALUE *pnop, VALUE child)
 {
     const char *types = insn_op_types(insn);
     char type = types[op_no];
@@ -1035,7 +1200,7 @@ insn_operand_intern(rb_iseq_t *iseq,
 	break;
 
       case TS_IC:
-	ret = rb_sprintf("<ic:%"PRIdPTRDIFF">", (struct iseq_inline_cache_entry *)op - iseq->ic_entries);
+	ret = rb_sprintf("<is:%"PRIdPTRDIFF">", (union iseq_inline_storage_entry *)op - iseq->is_entries);
 	break;
 
       case TS_CALLINFO:
@@ -1112,9 +1277,9 @@ rb_iseq_disasm_insn(VALUE ret, VALUE *iseq, size_t pos,
 
     for (j = 0; types[j]; j++) {
 	const char *types = insn_op_types(insn);
-	VALUE opstr = insn_operand_intern(iseqdat, insn, j, iseq[pos + j + 1],
-					  len, pos, &iseq[pos + j + 2],
-					  child);
+	VALUE opstr = rb_insn_operand_intern(iseqdat, insn, j, iseq[pos + j + 1],
+					     len, pos, &iseq[pos + j + 2],
+					     child);
 	rb_str_concat(str, opstr);
 
 	if (types[j + 1]) {
@@ -1235,10 +1400,11 @@ rb_iseq_disasm(VALUE self)
     if (tbl) {
 	rb_str_catf(str,
 		    "local table (size: %d, argc: %d "
-		    "[opts: %d, rest: %d, post: %d, block: %d] s%d)\n",
+		    "[opts: %d, rest: %d, post: %d, block: %d, keyword: %d@%d] s%d)\n",
 		    iseqdat->local_size, iseqdat->argc,
 		    iseqdat->arg_opts, iseqdat->arg_rest,
 		    iseqdat->arg_post_len, iseqdat->arg_block,
+		    iseqdat->arg_keywords, iseqdat->local_size-iseqdat->arg_keyword,
 		    iseqdat->arg_simple);
 
 	for (i = 0; i < iseqdat->local_table_size; i++) {
@@ -1289,6 +1455,63 @@ rb_iseq_disasm(VALUE self)
     return str;
 }
 
+/*
+ *  Returns the instruction sequence containing the given proc or method.
+ *
+ *  For example, using irb:
+ *
+ *	# a proc
+ *	> p = proc { num = 1 + 2 }
+ *	> RubyVM::InstructionSequence.of(p)
+ *	> #=> <RubyVM::InstructionSequence:block in irb_binding@(irb)>
+ *
+ *	# for a method
+ *	> def foo(bar); puts bar; end
+ *	> RubyVM::InstructionSequence.of(method(:foo))
+ *	> #=> <RubyVM::InstructionSequence:foo@(irb)>
+ *
+ *  Using ::compile_file:
+ *
+ *	# /tmp/iseq_of.rb
+ *	def hello
+ *	  puts "hello, world"
+ *	end
+ *
+ *	$a_global_proc = proc { str = 'a' + 'b' }
+ *
+ *	# in irb
+ *	> require '/tmp/iseq_of.rb'
+ *
+ *	# first the method hello
+ *	> RubyVM::InstructionSequence.of(method(:hello))
+ *	> #=> #<RubyVM::InstructionSequence:0x007fb73d7cb1d0>
+ *
+ *	# then the global proc
+ *	> RubyVM::InstructionSequence.of($a_global_proc)
+ *	> #=> #<RubyVM::InstructionSequence:0x007fb73d7caf78>
+ */
+static VALUE
+iseq_s_of(VALUE klass, VALUE body)
+{
+    VALUE ret = Qnil;
+    rb_iseq_t *iseq;
+
+    rb_secure(1);
+
+    if (rb_obj_is_proc(body)) {
+	rb_proc_t *proc;
+	GetProcPtr(body, proc);
+	iseq = proc->block.iseq;
+	if (RUBY_VM_NORMAL_ISEQ_P(iseq)) {
+	    ret = iseq->self;
+	}
+    }
+    else if ((iseq = rb_method_get_iseq(body)) != 0) {
+	ret = iseq->self;
+    }
+    return ret;
+}
+
 /*
  *  call-seq:
  *     InstructionSequence.disasm(body) -> str
@@ -1341,27 +1564,12 @@ rb_iseq_disasm(VALUE self)
  *    0012 leave
  *
  */
+
 static VALUE
 iseq_s_disasm(VALUE klass, VALUE body)
 {
-    VALUE ret = Qnil;
-    rb_iseq_t *iseq;
-
-    rb_secure(1);
-
-    if (rb_obj_is_proc(body)) {
-	rb_proc_t *proc;
-	GetProcPtr(body, proc);
-	iseq = proc->block.iseq;
-	if (RUBY_VM_NORMAL_ISEQ_P(iseq)) {
-	    ret = rb_iseq_disasm(iseq->self);
-	}
-    }
-    else if ((iseq = rb_method_get_iseq(body)) != 0) {
-	ret = rb_iseq_disasm(iseq->self);
-    }
-
-    return ret;
+    VALUE iseqval = iseq_s_of(klass, body);
+    return NIL_P(iseqval) ? Qnil : rb_iseq_disasm(iseqval);
 }
 
 const char *
@@ -1397,7 +1605,7 @@ static VALUE
 exception_type2symbol(VALUE type)
 {
     ID id;
-    switch(type) {
+    switch (type) {
       case CATCH_TYPE_RESCUE: CONST_ID(id, "rescue"); break;
       case CATCH_TYPE_ENSURE: CONST_ID(id, "ensure"); break;
       case CATCH_TYPE_RETRY:  CONST_ID(id, "retry");  break;
@@ -1466,7 +1674,7 @@ iseq_data_to_ary(rb_iseq_t *iseq)
     }
 
     /* type */
-    switch(iseq->type) {
+    switch (iseq->type) {
       case ISEQ_TYPE_TOP:    type = sym_top;    break;
       case ISEQ_TYPE_METHOD: type = sym_method; break;
       case ISEQ_TYPE_BLOCK:  type = sym_block;  break;
@@ -1565,10 +1773,11 @@ iseq_data_to_ary(rb_iseq_t *iseq)
 		    rb_ary_push(ary, ID2SYM(entry->id));
 		}
 		break;
-	      case TS_IC: {
-		  struct iseq_inline_cache_entry *ic = (struct iseq_inline_cache_entry *)*seq;
-		    rb_ary_push(ary, INT2FIX(ic - iseq->ic_entries));
-	        }
+	      case TS_IC:
+		{
+		    union iseq_inline_storage_entry *is = (union iseq_inline_storage_entry *)*seq;
+		    rb_ary_push(ary, INT2FIX(is - iseq->is_entries));
+		}
 		break;
 	      case TS_CALLINFO:
 		{
@@ -1636,14 +1845,14 @@ iseq_data_to_ary(rb_iseq_t *iseq)
     ti = 0;
 
     for (i=0, pos=0; i<RARRAY_LEN(nbody); i++) {
-	VALUE ary = RARRAY_PTR(nbody)[i];
+	VALUE ary = RARRAY_AREF(nbody, i);
 	st_data_t label;
 
 	if (st_lookup(labels_table, pos, &label)) {
 	    rb_ary_push(body, (VALUE)label);
 	}
 
-	if (iseq->line_info_size < ti && iseq->line_info_table[ti].position == pos) {
+	if (ti < iseq->line_info_size && iseq->line_info_table[ti].position == pos) {
 	    line = iseq->line_info_table[ti].line_no;
 	    rb_ary_push(body, INT2FIX(line));
 	    ti++;
@@ -1691,22 +1900,23 @@ rb_iseq_clone(VALUE iseqval, VALUE newcbase)
     GetISeqPtr(iseqval, iseq0);
     GetISeqPtr(newiseq, iseq1);
 
-    *iseq1 = *iseq0;
+    MEMCPY(iseq1, iseq0, rb_iseq_t, 1); /* TODO: write barrier? */
+
     iseq1->self = newiseq;
     if (!iseq1->orig) {
-	iseq1->orig = iseqval;
+	OBJ_WRITE(iseq1->self, &iseq1->orig, iseqval);
     }
     if (iseq0->local_iseq == iseq0) {
 	iseq1->local_iseq = iseq1;
     }
     if (newcbase) {
-	iseq1->cref_stack = NEW_CREF(newcbase);
+	ISEQ_SET_CREF(iseq1, NEW_CREF(newcbase));
 	iseq1->cref_stack->nd_refinements = iseq0->cref_stack->nd_refinements;
 	iseq1->cref_stack->nd_visi = iseq0->cref_stack->nd_visi;
 	if (iseq0->cref_stack->nd_next) {
 	    iseq1->cref_stack->nd_next = iseq0->cref_stack->nd_next;
 	}
-	iseq1->klass = newcbase;
+	OBJ_WRITE(iseq1, &iseq1->klass, newcbase);
     }
 
     return newiseq;
@@ -1740,12 +1950,7 @@ rb_iseq_parameters(const rb_iseq_t *iseq, int is_proc)
 	    rb_ary_push(args, PARAM(i, req));
 	}
     }
-    r = iseq->arg_rest != -1 ? iseq->arg_rest :
-	iseq->arg_post_len > 0 ? iseq->arg_post_start :
-	iseq->arg_block != -1 ? iseq->arg_block :
-	iseq->arg_keyword != -1 ? iseq->arg_keyword :
-	iseq->arg_size;
-    if (iseq->arg_keyword != -1) r -= iseq->arg_keywords;
+    r = iseq->argc + iseq->arg_opts - 1;
     for (; i < r; i++) {
 	PARAM_TYPE(opt);
 	if (rb_id2str(PARAM_ID(i))) {
@@ -1771,8 +1976,20 @@ rb_iseq_parameters(const rb_iseq_t *iseq, int is_proc)
 	}
     }
     if (iseq->arg_keyword != -1) {
+	i = 0;
+	if (iseq->arg_keyword_required) {
+	    ID keyreq;
+	    CONST_ID(keyreq, "keyreq");
+	    for (; i < iseq->arg_keyword_required; i++) {
+		PARAM_TYPE(keyreq);
+		if (rb_id2str(iseq->arg_keyword_table[i])) {
+		    rb_ary_push(a, ID2SYM(iseq->arg_keyword_table[i]));
+		}
+		rb_ary_push(args, a);
+	    }
+	}
 	CONST_ID(key, "key");
-	for (i = 0; i < iseq->arg_keywords; i++) {
+	for (; i < iseq->arg_keywords; i++) {
 	    PARAM_TYPE(key);
 	    if (rb_id2str(iseq->arg_keyword_table[i])) {
 		rb_ary_push(a, ID2SYM(iseq->arg_keyword_table[i]));
@@ -1850,12 +2067,11 @@ rb_iseq_build_for_ruby2cext(
     GetISeqPtr(iseqval, iseq);
 
     /* copy iseq */
-    *iseq = *iseq_template;
-    iseq->location.label = rb_str_new2(name);
-    iseq->location.path = rb_str_new2(path);
+    MEMCPY(iseq, iseq_template, rb_iseq_t, 1); /* TODO: write barrier, *iseq = *iseq_template; */
+    OBJ_WRITE(iseq->self, &iseq->location.label, rb_str_new2(name));
+    OBJ_WRITE(iseq->self, &iseq->location.path, rb_str_new2(path));
     iseq->location.first_lineno = first_lineno;
-    iseq->mark_ary = rb_ary_tmp_new(3);
-    OBJ_UNTRUST(iseq->mark_ary);
+    OBJ_WRITE(iseq->self, &iseq->mark_ary, 0);
     iseq->self = iseqval;
 
     iseq->iseq = ALLOC_N(VALUE, iseq->iseq_size);
@@ -1888,21 +2104,146 @@ rb_iseq_build_for_ruby2cext(
     return iseqval;
 }
 
+/* Experimental tracing support: trace(line) -> trace(specified_line)
+ * MRI Specific.
+ */
+
+int
+rb_iseq_line_trace_each(VALUE iseqval, int (*func)(int line, rb_event_flag_t *events_ptr, void *d), void *data)
+{
+    int trace_num = 0;
+    size_t pos, insn;
+    rb_iseq_t *iseq;
+    int cont = 1;
+    GetISeqPtr(iseqval, iseq);
+
+    for (pos = 0; cont && pos < iseq->iseq_size; pos += insn_len(insn)) {
+	insn = iseq->iseq[pos];
+
+	if (insn == BIN(trace)) {
+	    rb_event_flag_t current_events = (VALUE)iseq->iseq[pos+1];
+
+	    if (current_events & RUBY_EVENT_LINE) {
+		rb_event_flag_t events = current_events & RUBY_EVENT_SPECIFIED_LINE;
+		trace_num++;
+
+		if (func) {
+		    int line = find_line_no(iseq, pos);
+		    /* printf("line: %d\n", line); */
+		    cont = (*func)(line, &events, data);
+		    if (current_events != events) {
+			iseq->iseq[pos+1] = iseq->iseq_encoded[pos+1] =
+			  (VALUE)(current_events | (events & RUBY_EVENT_SPECIFIED_LINE));
+		    }
+		}
+	    }
+	}
+    }
+    return trace_num;
+}
+
+static int
+collect_trace(int line, rb_event_flag_t *events_ptr, void *ptr)
+{
+    VALUE result = (VALUE)ptr;
+    rb_ary_push(result, INT2NUM(line));
+    return 1;
+}
+
+/*
+ * <b>Experimental MRI specific feature, only available as C level api.</b>
+ *
+ * Returns all +specified_line+ events.
+ */
+VALUE
+rb_iseq_line_trace_all(VALUE iseqval)
+{
+    VALUE result = rb_ary_new();
+    rb_iseq_line_trace_each(iseqval, collect_trace, (void *)result);
+    return result;
+}
+
+struct set_specifc_data {
+    int pos;
+    int set;
+    int prev; /* 1: set, 2: unset, 0: not found */
+};
+
+static int
+line_trace_specify(int line, rb_event_flag_t *events_ptr, void *ptr)
+{
+    struct set_specifc_data *data = (struct set_specifc_data *)ptr;
+
+    if (data->pos == 0) {
+	data->prev = *events_ptr & RUBY_EVENT_SPECIFIED_LINE ? 1 : 2;
+	if (data->set) {
+	    *events_ptr = *events_ptr | RUBY_EVENT_SPECIFIED_LINE;
+	}
+	else {
+	    *events_ptr = *events_ptr & ~RUBY_EVENT_SPECIFIED_LINE;
+	}
+	return 0; /* found */
+    }
+    else {
+	data->pos--;
+	return 1;
+    }
+}
+
+/*
+ * <b>Experimental MRI specific feature, only available as C level api.</b>
+ *
+ * Set a +specified_line+ event at the given line position, if the +set+
+ * parameter is +true+.
+ *
+ * This method is useful for building a debugger breakpoint at a specific line.
+ *
+ * A TypeError is raised if +set+ is not boolean.
+ *
+ * If +pos+ is a negative integer a TypeError exception is raised.
+ */
+VALUE
+rb_iseq_line_trace_specify(VALUE iseqval, VALUE pos, VALUE set)
+{
+    struct set_specifc_data data;
+
+    data.prev = 0;
+    data.pos = NUM2INT(pos);
+    if (data.pos < 0) rb_raise(rb_eTypeError, "`pos' is negative");
+
+    switch (set) {
+      case Qtrue:  data.set = 1; break;
+      case Qfalse: data.set = 0; break;
+      default:
+	rb_raise(rb_eTypeError, "`set' should be true/false");
+    }
+
+    rb_iseq_line_trace_each(iseqval, line_trace_specify, (void *)&data);
+
+    if (data.prev == 0) {
+	rb_raise(rb_eTypeError, "`pos' is out of range.");
+    }
+    return data.prev == 1 ? Qtrue : Qfalse;
+}
+
 /*
  *  Document-class: RubyVM::InstructionSequence
  *
  *  The InstructionSequence class represents a compiled sequence of
- *  instructions for the Ruby Virtual Machine. With it, you can get a handle
- *  to the instructions that make up a method or a proc, compile strings of
- *  Ruby code down to VM instructions, and disassemble instruction sequences
- *  to strings for easy inspection. It is mostly useful if you want to learn
- *  how the Ruby VM works, but it also lets you control various settings for
- *  the Ruby iseq compiler. You can find the source for the VM instructions in
- *  +insns.def+ in the Ruby source.
+ *  instructions for the Ruby Virtual Machine.
  *
- *  The instruction sequence that results from a given piece of Ruby code will
- *  almost certainly change as Ruby changes, so example output in this
- *  documentation may be different from what you see.
+ *  With it, you can get a handle to the instructions that make up a method or
+ *  a proc, compile strings of Ruby code down to VM instructions, and
+ *  disassemble instruction sequences to strings for easy inspection. It is
+ *  mostly useful if you want to learn how the Ruby VM works, but it also lets
+ *  you control various settings for the Ruby iseq compiler.
+ *
+ *  You can find the source for the VM instructions in +insns.def+ in the Ruby
+ *  source.
+ *
+ *  The instruction sequence results will almost certainly change as Ruby
+ *  changes, so example output in this documentation may be different from what
+ *  you see.
  */
 
 void
@@ -1917,9 +2258,26 @@ Init_ISeq(void)
     rb_define_method(rb_cISeq, "to_a", iseq_to_a, 0);
     rb_define_method(rb_cISeq, "eval", iseq_eval, 0);
 
+    /* location APIs */
+    rb_define_method(rb_cISeq, "path", iseq_path, 0);
+    rb_define_method(rb_cISeq, "absolute_path", iseq_absolute_path, 0);
+    rb_define_method(rb_cISeq, "label", iseq_label, 0);
+    rb_define_method(rb_cISeq, "base_label", iseq_base_label, 0);
+    rb_define_method(rb_cISeq, "first_lineno", iseq_first_lineno, 0);
+
+#if 0
+    /* Now, it is experimental. No discussions, no tests. */
+    /* They can be used from C level. Please give us feedback. */
+    rb_define_method(rb_cISeq, "line_trace_all", rb_iseq_line_trace_all, 0);
+    rb_define_method(rb_cISeq, "line_trace_specify", rb_iseq_line_trace_specify, 2);
+#else
+    (void)rb_iseq_line_trace_all;
+    (void)rb_iseq_line_trace_specify;
+#endif
+
 #if 0 /* TBD */
-    rb_define_method(rb_cISeq, "marshal_dump", iseq_marshal_dump, 0);
-    rb_define_method(rb_cISeq, "marshal_load", iseq_marshal_load, 1);
+    rb_define_private_method(rb_cISeq, "marshal_dump", iseq_marshal_dump, 0);
+    rb_define_private_method(rb_cISeq, "marshal_load", iseq_marshal_load, 1);
 #endif
 
     /* disable this feature because there is no verifier. */
@@ -1933,4 +2291,5 @@ Init_ISeq(void)
     rb_define_singleton_method(rb_cISeq, "compile_option=", iseq_s_compile_option_set, 1);
     rb_define_singleton_method(rb_cISeq, "disasm", iseq_s_disasm, 1);
     rb_define_singleton_method(rb_cISeq, "disassemble", iseq_s_disasm, 1);
+    rb_define_singleton_method(rb_cISeq, "of", iseq_s_of, 1);
 }
