@@ -2,7 +2,7 @@
 
   marshal.c -
 
-  $Author: nari $
+  $Author: akr $
   created at: Thu Apr 27 16:30:01 JST 1995
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -42,7 +42,7 @@ shortlen(long len, BDIGIT *ds)
 	num = SHORTDN(num);
 	offset++;
     }
-    return (len - 1)*sizeof(BDIGIT)/2 + offset;
+    return (len - 1)*SIZEOF_BDIGITS/2 + offset;
 }
 #define SHORTLEN(x) shortlen((x),d)
 #endif
@@ -130,7 +130,7 @@ rb_marshal_define_compat(VALUE newclass, VALUE oldclass, VALUE (*dumper)(VALUE),
     st_insert(compat_allocator_tbl, (st_data_t)allocator, (st_data_t)compat);
 }
 
-#define MARSHAL_INFECTION (FL_TAINT|FL_UNTRUSTED)
+#define MARSHAL_INFECTION FL_TAINT
 typedef char ruby_check_marshal_viral_flags[MARSHAL_INFECTION == (int)MARSHAL_INFECTION ? 1 : -1];
 
 struct dump_arg {
@@ -457,12 +457,17 @@ hash_each(VALUE key, VALUE value, struct dump_call_arg *arg)
     return ST_CONTINUE;
 }
 
+#define SINGLETON_DUMP_UNABLE_P(klass) \
+    (RCLASS_M_TBL(klass)->num_entries || \
+     (RCLASS_IV_TBL(klass) && RCLASS_IV_TBL(klass)->num_entries > 1))
+
 static void
 w_extended(VALUE klass, struct dump_arg *arg, int check)
 {
     if (check && FL_TEST(klass, FL_SINGLETON)) {
-	if (RCLASS_M_TBL(klass)->num_entries ||
-	    (RCLASS_IV_TBL(klass) && RCLASS_IV_TBL(klass)->num_entries > 1)) {
+	VALUE origin = RCLASS_ORIGIN(klass);
+	if (SINGLETON_DUMP_UNABLE_P(klass) ||
+	    (origin != klass && SINGLETON_DUMP_UNABLE_P(origin))) {
 	    rb_raise(rb_eTypeError, "singleton can't be dumped");
 	}
 	klass = RCLASS_SUPER(klass);
@@ -593,7 +598,7 @@ w_object(VALUE obj, struct dump_arg *arg, int limit)
     st_table *ivtbl = 0;
     st_data_t num;
     int hasiv = 0;
-#define has_ivars(obj, ivtbl) (((ivtbl) = rb_generic_ivar_table(obj)) != 0 || \
+#define has_ivars(obj, ivtbl) ((((ivtbl) = rb_generic_ivar_table(obj)) != 0) || \
 			       (!SPECIAL_CONST_P(obj) && !ENCODING_IS_ASCII8BIT(obj)))
 
     if (limit == 0) {
@@ -637,32 +642,30 @@ w_object(VALUE obj, struct dump_arg *arg, int limit)
 	w_symbol(SYM2ID(obj), arg);
     }
     else if (FLONUM_P(obj)) {
+	st_add_direct(arg->data, obj, arg->data->num_entries);
 	w_byte(TYPE_FLOAT, arg);
 	w_float(RFLOAT_VALUE(obj), arg);
     }
     else {
+	VALUE v;
+
 	arg->infection |= (int)FL_TEST(obj, MARSHAL_INFECTION);
 
-	if (rb_respond_to(obj, s_mdump)) {
-	    volatile VALUE v;
+	if (rb_obj_respond_to(obj, s_mdump, TRUE)) {
+	    st_add_direct(arg->data, obj, arg->data->num_entries);
 
-            st_add_direct(arg->data, obj, arg->data->num_entries);
-
-	    v = rb_funcall(obj, s_mdump, 0, 0);
+	    v = rb_funcall2(obj, s_mdump, 0, 0);
 	    check_dump_arg(arg, s_mdump);
-	    hasiv = has_ivars(obj, ivtbl);
-	    if (hasiv) w_byte(TYPE_IVAR, arg);
 	    w_class(TYPE_USRMARSHAL, obj, arg, FALSE);
 	    w_object(v, arg, limit);
-	    if (hasiv) w_ivar(obj, ivtbl, &c_arg);
 	    return;
 	}
-	if (rb_respond_to(obj, s_dump)) {
-	    VALUE v;
+	if (rb_obj_respond_to(obj, s_dump, TRUE)) {
             st_table *ivtbl2 = 0;
             int hasiv2;
 
-	    v = rb_funcall(obj, s_dump, 1, INT2NUM(limit));
+	    v = INT2NUM(limit);
+	    v = rb_funcall2(obj, s_dump, 1, &v);
 	    check_dump_arg(arg, s_dump);
 	    if (!RB_TYPE_P(v, T_STRING)) {
 		rb_raise(rb_eTypeError, "_dump() must return string");
@@ -709,8 +712,9 @@ w_object(VALUE obj, struct dump_arg *arg, int limit)
 	    }
 	    w_byte(TYPE_CLASS, arg);
 	    {
-		volatile VALUE path = class2path(obj);
+		VALUE path = class2path(obj);
 		w_bytes(RSTRING_PTR(path), RSTRING_LEN(path), arg);
+		RB_GC_GUARD(path);
 	    }
 	    break;
 
@@ -719,6 +723,7 @@ w_object(VALUE obj, struct dump_arg *arg, int limit)
 	    {
 		VALUE path = class2path(obj);
 		w_bytes(RSTRING_PTR(path), RSTRING_LEN(path), arg);
+		RB_GC_GUARD(path);
 	    }
 	    break;
 
@@ -778,7 +783,7 @@ w_object(VALUE obj, struct dump_arg *arg, int limit)
 
 		w_long(len, arg);
 		for (i=0; i<RARRAY_LEN(obj); i++) {
-		    w_object(RARRAY_PTR(obj)[i], arg, limit);
+		    w_object(RARRAY_AREF(obj, i), arg, limit);
 		    if (len != RARRAY_LEN(obj)) {
 			rb_raise(rb_eRuntimeError, "array modified during dump");
 		    }
@@ -815,8 +820,8 @@ w_object(VALUE obj, struct dump_arg *arg, int limit)
 		w_long(len, arg);
 		mem = rb_struct_members(obj);
 		for (i=0; i<len; i++) {
-		    w_symbol(SYM2ID(RARRAY_PTR(mem)[i]), arg);
-		    w_object(RSTRUCT_PTR(obj)[i], arg, limit);
+		    w_symbol(SYM2ID(RARRAY_AREF(mem, i)), arg);
+		    w_object(RSTRUCT_GET(obj, i), arg, limit);
 		}
 	    }
 	    break;
@@ -830,12 +835,12 @@ w_object(VALUE obj, struct dump_arg *arg, int limit)
 	    {
 		VALUE v;
 
-		if (!rb_respond_to(obj, s_dump_data)) {
+		if (!rb_obj_respond_to(obj, s_dump_data, TRUE)) {
 		    rb_raise(rb_eTypeError,
 			     "no _dump_data is defined for class %s",
 			     rb_obj_classname(obj));
 		}
-		v = rb_funcall(obj, s_dump_data, 0);
+		v = rb_funcall2(obj, s_dump_data, 0, 0);
 		check_dump_arg(arg, s_dump_data);
 		w_class(TYPE_DATA, obj, arg, TRUE);
 		w_object(v, arg, limit);
@@ -847,6 +852,7 @@ w_object(VALUE obj, struct dump_arg *arg, int limit)
 		     rb_obj_classname(obj));
 	    break;
 	}
+	RB_GC_GUARD(obj);
     }
     if (hasiv) {
 	w_ivar(obj, ivtbl, &c_arg);
@@ -930,7 +936,7 @@ marshal_dump(int argc, VALUE *argv)
 	else if (NIL_P(a1)) io_needed();
 	else port = a1;
     }
-    wrapper = TypedData_Make_Struct(rb_cData, struct dump_arg, &dump_arg_data, arg);
+    RB_GC_GUARD(wrapper) = TypedData_Make_Struct(rb_cData, struct dump_arg, &dump_arg_data, arg);
     arg->dest = 0;
     arg->symbols = st_init_numtable();
     arg->data    = st_init_numtable();
@@ -943,8 +949,7 @@ marshal_dump(int argc, VALUE *argv)
 	    io_needed();
 	}
 	arg->dest = port;
-	if (rb_respond_to(port, s_binmode)) {
-	    rb_funcall2(port, s_binmode, 0, 0);
+	if (rb_check_funcall(port, s_binmode, 0, 0) != Qundef) {
 	    check_dump_arg(arg, s_binmode);
 	}
     }
@@ -968,6 +973,9 @@ marshal_dump(int argc, VALUE *argv)
 
 struct load_arg {
     VALUE src;
+    char *buf;
+    long buflen;
+    long readable;
     long offset;
     st_table *symbols;
     st_table *data;
@@ -1021,6 +1029,13 @@ static VALUE r_object(struct load_arg *arg);
 static ID r_symbol(struct load_arg *arg);
 static VALUE path2class(VALUE path);
 
+NORETURN(static void too_short(void));
+static void
+too_short(void)
+{
+    rb_raise(rb_eArgError, "marshal data too short");
+}
+
 static st_index_t
 r_prepare(struct load_arg *arg)
 {
@@ -1030,6 +1045,27 @@ r_prepare(struct load_arg *arg)
     return idx;
 }
 
+static unsigned char
+r_byte1_buffered(struct load_arg *arg)
+{
+    if (arg->buflen == 0) {
+	long readable = arg->readable < BUFSIZ ? arg->readable : BUFSIZ;
+	VALUE str, n = LONG2NUM(readable);
+
+	str = rb_funcall2(arg->src, s_read, 1, &n);
+
+	check_load_arg(arg, s_read);
+	if (NIL_P(str)) too_short();
+	StringValue(str);
+	arg->infection |= (int)FL_TEST(str, MARSHAL_INFECTION);
+	memcpy(arg->buf, RSTRING_PTR(str), RSTRING_LEN(str));
+	arg->offset = 0;
+	arg->buflen = RSTRING_LEN(str);
+    }
+    arg->buflen--;
+    return arg->buf[arg->offset++];
+}
+
 static int
 r_byte(struct load_arg *arg)
 {
@@ -1040,15 +1076,19 @@ r_byte(struct load_arg *arg)
 	    c = (unsigned char)RSTRING_PTR(arg->src)[arg->offset++];
 	}
 	else {
-	    rb_raise(rb_eArgError, "marshal data too short");
+	    too_short();
 	}
     }
     else {
-	VALUE src = arg->src;
-	VALUE v = rb_funcall2(src, s_getbyte, 0, 0);
-	check_load_arg(arg, s_getbyte);
-	if (NIL_P(v)) rb_eof_error();
-	c = (unsigned char)NUM2CHR(v);
+	if (arg->readable >0 || arg->buflen > 0) {
+	    c = r_byte1_buffered(arg);
+	}
+	else {
+	    VALUE v = rb_funcall2(arg->src, s_getbyte, 0, 0);
+	    check_load_arg(arg, s_getbyte);
+	    if (NIL_P(v)) rb_eof_error();
+	    c = (unsigned char)NUM2CHR(v);
+	}
     }
     return c;
 }
@@ -1101,6 +1141,68 @@ r_long(struct load_arg *arg)
     return x;
 }
 
+static VALUE
+r_bytes1(long len, struct load_arg *arg)
+{
+    VALUE str, n = LONG2NUM(len);
+
+    str = rb_funcall2(arg->src, s_read, 1, &n);
+    check_load_arg(arg, s_read);
+    if (NIL_P(str)) too_short();
+    StringValue(str);
+    if (RSTRING_LEN(str) != len) too_short();
+    arg->infection |= (int)FL_TEST(str, MARSHAL_INFECTION);
+
+    return str;
+}
+
+static VALUE
+r_bytes1_buffered(long len, struct load_arg *arg)
+{
+    VALUE str;
+
+    if (len <= arg->buflen) {
+	str = rb_str_new(arg->buf+arg->offset, len);
+	arg->offset += len;
+	arg->buflen -= len;
+    }
+    else {
+	long buflen = arg->buflen;
+	long readable = arg->readable + 1;
+	long tmp_len, read_len, need_len = len - buflen;
+	VALUE tmp, n;
+
+	readable = readable < BUFSIZ ? readable : BUFSIZ;
+	read_len = need_len > readable ? need_len : readable;
+	n = LONG2NUM(read_len);
+	tmp = rb_funcall2(arg->src, s_read, 1, &n);
+
+	check_load_arg(arg, s_read);
+	if (NIL_P(tmp)) too_short();
+	StringValue(tmp);
+
+	tmp_len = RSTRING_LEN(tmp);
+
+	if (tmp_len < need_len) too_short();
+	arg->infection |= (int)FL_TEST(tmp, MARSHAL_INFECTION);
+
+	str = rb_str_new(arg->buf+arg->offset, buflen);
+	rb_str_cat(str, RSTRING_PTR(tmp), need_len);
+
+	if (tmp_len > need_len) {
+	    buflen = tmp_len - need_len;
+	    memcpy(arg->buf, RSTRING_PTR(tmp)+need_len, buflen);
+	    arg->buflen = buflen;
+	}
+	else {
+	    arg->buflen = 0;
+	}
+	arg->offset = 0;
+    }
+
+    return str;
+}
+
 #define r_bytes(arg) r_bytes0(r_long(arg), (arg))
 
 static VALUE
@@ -1115,19 +1217,16 @@ r_bytes0(long len, struct load_arg *arg)
 	    arg->offset += len;
 	}
 	else {
-	  too_short:
-	    rb_raise(rb_eArgError, "marshal data too short");
+	    too_short();
 	}
     }
     else {
-	VALUE src = arg->src;
-	VALUE n = LONG2NUM(len);
-	str = rb_funcall2(src, s_read, 1, &n);
-	check_load_arg(arg, s_read);
-	if (NIL_P(str)) goto too_short;
-	StringValue(str);
-	if (RSTRING_LEN(str) != len) goto too_short;
-	arg->infection |= (int)FL_TEST(str, MARSHAL_INFECTION);
+	if (arg->readable > 0 || arg->buflen > 0) {
+	    str = r_bytes1_buffered(len, arg);
+	}
+	else {
+	    str = r_bytes1(len, arg);
+	}
     }
     return str;
 }
@@ -1226,7 +1325,8 @@ r_entry0(VALUE v, st_index_t num, struct load_arg *arg)
     else {
         st_insert(arg->data, num, (st_data_t)v);
     }
-    if (arg->infection) {
+    if (arg->infection &&
+	!RB_TYPE_P(v, T_CLASS) && !RB_TYPE_P(v, T_MODULE)) {
 	FL_SET(v, arg->infection);
 	if ((VALUE)real_obj != Qundef)
 	    FL_SET((VALUE)real_obj, arg->infection);
@@ -1235,7 +1335,7 @@ r_entry0(VALUE v, st_index_t num, struct load_arg *arg)
 }
 
 static VALUE
-r_leave(VALUE v, struct load_arg *arg)
+r_fixup_compat(VALUE v, struct load_arg *arg)
 {
     st_data_t data;
     if (st_lookup(arg->compat_tbl, v, &data)) {
@@ -1249,6 +1349,12 @@ r_leave(VALUE v, struct load_arg *arg)
         st_delete(arg->compat_tbl, &key, 0);
         v = real_obj;
     }
+    return v;
+}
+
+static VALUE
+r_post_proc(VALUE v, struct load_arg *arg)
+{
     if (arg->proc) {
 	v = rb_funcall(arg->proc, s_call, 1, v);
 	check_load_arg(arg, s_call);
@@ -1256,6 +1362,32 @@ r_leave(VALUE v, struct load_arg *arg)
     return v;
 }
 
+static VALUE
+r_leave(VALUE v, struct load_arg *arg)
+{
+    v = r_fixup_compat(v, arg);
+    v = r_post_proc(v, arg);
+    return v;
+}
+
+static int
+copy_ivar_i(st_data_t key, st_data_t val, st_data_t arg)
+{
+    VALUE obj = (VALUE)arg, value = (VALUE)val;
+    ID vid = (ID)key;
+
+    if (!rb_ivar_defined(obj, vid))
+	rb_ivar_set(obj, vid, value);
+    return ST_CONTINUE;
+}
+
+static VALUE
+r_copy_ivar(VALUE v, VALUE data)
+{
+    rb_ivar_foreach(data, copy_ivar_i, (st_data_t)v);
+    return v;
+}
+
 static void
 r_ivar(VALUE obj, int *has_encoding, struct load_arg *arg)
 {
@@ -1284,20 +1416,18 @@ path2class(VALUE path)
     VALUE v = rb_path_to_class(path);
 
     if (!RB_TYPE_P(v, T_CLASS)) {
-	rb_raise(rb_eArgError, "%.*s does not refer to class",
-		 (int)RSTRING_LEN(path), RSTRING_PTR(path));
+	rb_raise(rb_eArgError, "%"PRIsVALUE" does not refer to class", path);
     }
     return v;
 }
 
-static VALUE
-path2module(VALUE path)
-{
-    VALUE v = rb_path_to_class(path);
+#define path2module(path) must_be_module(rb_path_to_class(path), path)
 
+static VALUE
+must_be_module(VALUE v, VALUE path)
+{
     if (!RB_TYPE_P(v, T_MODULE)) {
-	rb_raise(rb_eArgError, "%.*s does not refer to module",
-		 (int)RSTRING_LEN(path), RSTRING_PTR(path));
+	rb_raise(rb_eArgError, "%"PRIsVALUE" does not refer to module", path);
     }
     return v;
 }
@@ -1332,12 +1462,19 @@ append_extmod(VALUE obj, VALUE extmod)
 {
     long i = RARRAY_LEN(extmod);
     while (i > 0) {
-	VALUE m = RARRAY_PTR(extmod)[--i];
+	VALUE m = RARRAY_AREF(extmod, --i);
 	rb_extend_object(obj, m);
     }
     return obj;
 }
 
+#define prohibit_ivar(type, str) do { \
+	if (!ivp || !*ivp) break; \
+	rb_raise(rb_eTypeError, \
+		 "can't override instance variable of "type" `%"PRIsVALUE"'", \
+		 (str)); \
+    } while (0)
+
 static VALUE
 r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 {
@@ -1353,10 +1490,7 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 	    rb_raise(rb_eArgError, "dump format error (unlinked)");
 	}
 	v = (VALUE)link;
-	if (arg->proc) {
-	    v = rb_funcall(arg->proc, s_call, 1, v);
-	    check_load_arg(arg, s_call);
-	}
+	r_post_proc(v, arg);
 	break;
 
       case TYPE_IVAR:
@@ -1370,16 +1504,36 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
       case TYPE_EXTENDED:
 	{
-	    VALUE m = path2module(r_unique(arg));
+	    VALUE path = r_unique(arg);
+	    VALUE m = rb_path_to_class(path);
 
-            if (NIL_P(extmod)) extmod = rb_ary_tmp_new(0);
-            rb_ary_push(extmod, m);
+	    if (RB_TYPE_P(m, T_CLASS)) { /* prepended */
+		VALUE c;
 
-	    v = r_object0(arg, 0, extmod);
-            while (RARRAY_LEN(extmod) > 0) {
-                m = rb_ary_pop(extmod);
-                rb_extend_object(v, m);
-            }
+		v = r_object0(arg, 0, Qnil);
+		c = CLASS_OF(v);
+		if (c != m || FL_TEST(c, FL_SINGLETON)) {
+		    rb_raise(rb_eArgError,
+			     "prepended class %"PRIsVALUE" differs from class %"PRIsVALUE,
+			     path, rb_class_name(c));
+		}
+		c = rb_singleton_class(v);
+		while (RARRAY_LEN(extmod) > 0) {
+		    m = rb_ary_pop(extmod);
+		    rb_prepend_module(c, m);
+		}
+	    }
+	    else {
+		must_be_module(m, path);
+		if (NIL_P(extmod)) extmod = rb_ary_tmp_new(0);
+		rb_ary_push(extmod, m);
+
+		v = r_object0(arg, 0, extmod);
+		while (RARRAY_LEN(extmod) > 0) {
+		    m = rb_ary_pop(extmod);
+		    rb_extend_object(v, m);
+		}
+	    }
 	}
 	break;
 
@@ -1400,7 +1554,7 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
 		if (TYPE(v) != TYPE(tmp)) goto format_error;
 	    }
-	    RBASIC(v)->klass = c;
+	    RBASIC_SET_CLASS(v, c);
 	}
 	break;
 
@@ -1456,43 +1610,15 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
       case TYPE_BIGNUM:
 	{
 	    long len;
-	    BDIGIT *digits;
-	    volatile VALUE data;
+	    VALUE data;
+            int sign;
 
-	    NEWOBJ_OF(big, struct RBignum, rb_cBignum, T_BIGNUM);
-	    RBIGNUM_SET_SIGN(big, (r_byte(arg) == '+'));
+	    sign = r_byte(arg);
 	    len = r_long(arg);
 	    data = r_bytes0(len * 2, arg);
-#if SIZEOF_BDIGITS == SIZEOF_SHORT
-            rb_big_resize((VALUE)big, len);
-#else
-            rb_big_resize((VALUE)big, (len + 1) * 2 / sizeof(BDIGIT));
-#endif
-            digits = RBIGNUM_DIGITS(big);
-	    MEMCPY(digits, RSTRING_PTR(data), char, len * 2);
-#if SIZEOF_BDIGITS > SIZEOF_SHORT
-	    MEMZERO((char *)digits + len * 2, char,
-		    RBIGNUM_LEN(big) * sizeof(BDIGIT) - len * 2);
-#endif
-	    len = RBIGNUM_LEN(big);
-	    while (len > 0) {
-		unsigned char *p = (unsigned char *)digits;
-		BDIGIT num = 0;
-#if SIZEOF_BDIGITS > SIZEOF_SHORT
-		int shift = 0;
-		int i;
-
-		for (i=0; i<SIZEOF_BDIGITS; i++) {
-		    num |= (int)p[i] << shift;
-		    shift += 8;
-		}
-#else
-		num = p[0] | (p[1] << 8);
-#endif
-		*digits++ = num;
-		len--;
-	    }
-	    v = rb_big_norm((VALUE)big);
+            v = rb_integer_unpack(RSTRING_PTR(data), len, 2, 0,
+                INTEGER_PACK_LITTLE_ENDIAN | (sign == '-' ? INTEGER_PACK_NEGATIVE : 0));
+	    rb_str_resize(data, 0L);
 	    v = r_entry(v, arg);
             v = r_leave(v, arg);
 	}
@@ -1505,7 +1631,7 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
       case TYPE_REGEXP:
 	{
-	    volatile VALUE str = r_bytes(arg);
+	    VALUE str = r_bytes(arg);
 	    int options = r_byte(arg);
 	    int has_encoding = FALSE;
 	    st_index_t idx = r_prepare(arg);
@@ -1544,10 +1670,13 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
 	    v = rb_ary_new2(len);
 	    v = r_entry(v, arg);
+	    arg->readable += len - 1;
 	    while (len--) {
 		rb_ary_push(v, r_object(arg));
+		arg->readable--;
 	    }
             v = r_leave(v, arg);
+	    arg->readable++;
 	}
 	break;
 
@@ -1558,13 +1687,16 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
 	    v = rb_hash_new();
 	    v = r_entry(v, arg);
+	    arg->readable += (len - 1) * 2;
 	    while (len--) {
 		VALUE key = r_object(arg);
 		VALUE value = r_object(arg);
 		rb_hash_aset(v, key, value);
+		arg->readable -= 2;
 	    }
+	    arg->readable += 2;
 	    if (type == TYPE_HASH_DEF) {
-		RHASH_IFNONE(v) = r_object(arg);
+		RHASH_SET_IFNONE(v, r_object(arg));
 	    }
             v = r_leave(v, arg);
 	}
@@ -1589,21 +1721,24 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
                          rb_class2name(klass));
             }
 
+	    arg->readable += (len - 1) * 2;
 	    v = r_entry0(v, idx, arg);
 	    values = rb_ary_new2(len);
 	    for (i=0; i<len; i++) {
 		slot = r_symbol(arg);
 
-		if (RARRAY_PTR(mem)[i] != ID2SYM(slot)) {
+		if (RARRAY_AREF(mem, i) != ID2SYM(slot)) {
 		    rb_raise(rb_eTypeError, "struct %s not compatible (:%s for :%s)",
 			     rb_class2name(klass),
 			     rb_id2name(slot),
-			     rb_id2name(SYM2ID(RARRAY_PTR(mem)[i])));
+			     rb_id2name(SYM2ID(RARRAY_AREF(mem, i))));
 		}
                 rb_ary_push(values, r_object(arg));
+		arg->readable -= 2;
 	    }
             rb_struct_initialize(v, values);
             v = r_leave(v, arg);
+	    arg->readable += 2;
 	}
 	break;
 
@@ -1612,7 +1747,7 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 	    VALUE klass = path2class(r_unique(arg));
 	    VALUE data;
 
-	    if (!rb_respond_to(klass, s_load)) {
+	    if (!rb_obj_respond_to(klass, s_load, TRUE)) {
 		rb_raise(rb_eTypeError, "class %s needs to have method `_load'",
 			 rb_class2name(klass));
 	    }
@@ -1621,7 +1756,7 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 		r_ivar(data, NULL, arg);
 		*ivp = FALSE;
 	    }
-	    v = rb_funcall(klass, s_load, 1, data);
+	    v = rb_funcall2(klass, s_load, 1, &data);
 	    check_load_arg(arg, s_load);
 	    v = r_entry(v, arg);
             v = r_leave(v, arg);
@@ -1639,15 +1774,17 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 		/* for the case marshal_load is overridden */
 		append_extmod(v, extmod);
             }
-	    if (!rb_respond_to(v, s_mload)) {
+	    if (!rb_obj_respond_to(v, s_mload, TRUE)) {
 		rb_raise(rb_eTypeError, "instance of %s needs to have method `marshal_load'",
 			 rb_class2name(klass));
 	    }
 	    v = r_entry(v, arg);
 	    data = r_object(arg);
-	    rb_funcall(v, s_mload, 1, data);
+	    rb_funcall2(v, s_mload, 1, &data);
 	    check_load_arg(arg, s_mload);
-            v = r_leave(v, arg);
+	    v = r_fixup_compat(v, arg);
+	    v = r_copy_ivar(v, data);
+	    v = r_post_proc(v, arg);
 	    if (!NIL_P(extmod)) {
 		if (oldclass) append_extmod(v, extmod);
 		rb_ary_clear(extmod);
@@ -1672,18 +1809,20 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 	{
 	    VALUE klass = path2class(r_unique(arg));
 	    VALUE oldclass = 0;
+	    VALUE r;
 
 	    v = obj_alloc_by_klass(klass, arg, &oldclass);
 	    if (!RB_TYPE_P(v, T_DATA)) {
 		rb_raise(rb_eArgError, "dump format error");
 	    }
 	    v = r_entry(v, arg);
-	    if (!rb_respond_to(v, s_load_data)) {
+	    if (!rb_obj_respond_to(v, s_load_data, TRUE)) {
 		rb_raise(rb_eTypeError,
 			 "class %s needs to have instance method `_load_data'",
 			 rb_class2name(klass));
 	    }
-	    rb_funcall(v, s_load_data, 1, r_object0(arg, 0, extmod));
+	    r = r_object0(arg, 0, extmod);
+	    rb_funcall2(v, s_load_data, 1, &r);
 	    check_load_arg(arg, s_load_data);
 	    v = r_leave(v, arg);
 	}
@@ -1691,9 +1830,10 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
       case TYPE_MODULE_OLD:
         {
-	    volatile VALUE str = r_bytes(arg);
+	    VALUE str = r_bytes(arg);
 
 	    v = rb_path_to_class(str);
+	    prohibit_ivar("class/module", str);
 	    v = r_entry(v, arg);
             v = r_leave(v, arg);
 	}
@@ -1701,9 +1841,10 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
       case TYPE_CLASS:
         {
-	    volatile VALUE str = r_bytes(arg);
+	    VALUE str = r_bytes(arg);
 
 	    v = path2class(str);
+	    prohibit_ivar("class", str);
 	    v = r_entry(v, arg);
             v = r_leave(v, arg);
 	}
@@ -1711,9 +1852,10 @@ r_object0(struct load_arg *arg, int *ivp, VALUE extmod)
 
       case TYPE_MODULE:
         {
-	    volatile VALUE str = r_bytes(arg);
+	    VALUE str = r_bytes(arg);
 
 	    v = path2module(str);
+	    prohibit_ivar("module", str);
 	    v = r_entry(v, arg);
             v = r_leave(v, arg);
 	}
@@ -1750,6 +1892,13 @@ r_object(struct load_arg *arg)
 static void
 clear_load_arg(struct load_arg *arg)
 {
+    if (arg->buf) {
+	xfree(arg->buf);
+	arg->buf = 0;
+    }
+    arg->buflen = 0;
+    arg->offset = 0;
+    arg->readable = 0;
     if (!arg->symbols) return;
     st_free_table(arg->symbols);
     arg->symbols = 0;
@@ -1769,6 +1918,9 @@ clear_load_arg(struct load_arg *arg)
  * may be either an instance of IO or an object that responds to
  * to_str. If proc is specified, it will be passed each object as it
  * is deserialized.
+ *
+ * Never pass untrusted data (including user supplied input) to this method.
+ * Please see the overview for further details.
  */
 static VALUE
 marshal_load(int argc, VALUE *argv)
@@ -1786,15 +1938,13 @@ marshal_load(int argc, VALUE *argv)
 	port = v;
     }
     else if (rb_respond_to(port, s_getbyte) && rb_respond_to(port, s_read)) {
-	if (rb_respond_to(port, s_binmode)) {
-	    rb_funcall2(port, s_binmode, 0, 0);
-	}
-	infection = (int)(FL_TAINT | FL_TEST(port, FL_UNTRUSTED));
+	rb_check_funcall(port, s_binmode, 0, 0);
+	infection = (int)FL_TAINT;
     }
     else {
 	io_needed();
     }
-    wrapper = TypedData_Make_Struct(rb_cData, struct load_arg, &load_arg_data, arg);
+    RB_GC_GUARD(wrapper) = TypedData_Make_Struct(rb_cData, struct load_arg, &load_arg_data, arg);
     arg->infection = infection;
     arg->src = port;
     arg->offset = 0;
@@ -1802,6 +1952,12 @@ marshal_load(int argc, VALUE *argv)
     arg->data    = st_init_numtable();
     arg->compat_tbl = st_init_numtable();
     arg->proc = 0;
+    arg->readable = 0;
+
+    if (NIL_P(v))
+	arg->buf = xmalloc(BUFSIZ);
+    else
+	arg->buf = 0;
 
     major = r_byte(arg);
     minor = r_byte(arg);
@@ -1859,6 +2015,21 @@ marshal_load(int argc, VALUE *argv)
  * precedence over _dump if both are defined.  marshal_dump may result in
  * smaller Marshal strings.
  *
+ * == Security considerations
+ *
+ * By design, Marshal.load can deserialize almost any class loaded into the
+ * Ruby process. In many cases this can lead to remote code execution if the
+ * Marshal data is loaded from an untrusted source.
+ *
+ * As a result, Marshal.load is not suitable as a general purpose serialization
+ * format and you should never unmarshal user supplied input or other untrusted
+ * data.
+ *
+ * If you need to deserialize untrusted data, use JSON or another serialization
+ * format that is only able to load simple, 'primitive' types such as String,
+ * Array, Hash, etc. Never allow user input to specify arbitrary types to
+ * deserialize into.
+ *
  * == marshal_dump and marshal_load
  *
  * When dumping an object the method marshal_dump will be called.
@@ -1946,7 +2117,9 @@ Init_marshal(void)
     rb_define_module_function(rb_mMarshal, "load", marshal_load, -1);
     rb_define_module_function(rb_mMarshal, "restore", marshal_load, -1);
 
+    /* major version */
     rb_define_const(rb_mMarshal, "MAJOR_VERSION", INT2FIX(MARSHAL_MAJOR));
+    /* minor version */
     rb_define_const(rb_mMarshal, "MINOR_VERSION", INT2FIX(MARSHAL_MINOR));
 
     compat_allocator_tbl = st_init_numtable();
