@@ -1,9 +1,9 @@
-# $Id: test_http.rb 37056 2012-09-29 11:19:06Z shirosaki $
-
+# coding: US-ASCII
 require 'test/unit'
 require 'net/http'
 require 'stringio'
 require_relative 'utils'
+require_relative '../../ruby/envutil'
 
 class TestNetHTTP < Test::Unit::TestCase
 
@@ -232,6 +232,8 @@ module TestNetHTTP_version_1_1_methods
     assert_nothing_raised {
       http.get('/', { 'User-Agent' => 'test' }.freeze)
     }
+
+    assert res.decode_content, '[Bug #7924]' if Net::HTTP::HAVE_ZLIB
   end
 
   def _test_get__iter(http)
@@ -292,8 +294,11 @@ module TestNetHTTP_version_1_1_methods
   def test_get2
     start {|http|
       http.get2('/') {|res|
-        assert_kind_of Net::HTTPResponse, res
-        assert_kind_of Net::HTTPResponse, res.header
+        EnvUtil.suppress_warning do
+          assert_kind_of Net::HTTPResponse, res
+          assert_kind_of Net::HTTPResponse, res.header
+        end
+
         unless self.is_a?(TestNetHTTP_v1_2_chunked)
           assert_not_nil res['content-length']
         end
@@ -311,6 +316,7 @@ module TestNetHTTP_version_1_1_methods
     start {|http|
       _test_post__base http
       _test_post__file http
+      _test_post__no_data http
     }
   end
 
@@ -333,6 +339,16 @@ module TestNetHTTP_version_1_1_methods
     assert_equal data, f.string
   end
 
+  def _test_post__no_data(http)
+    unless self.is_a?(TestNetHTTP_v1_2_chunked)
+      EnvUtil.suppress_warning do
+        data = nil
+        res = http.post('/', data)
+        assert_not_equal '411', res.code
+      end
+    end
+  end
+
   def test_s_post_form
     url = "http://#{config('host')}:#{config('port')}/"
     res = Net::HTTP.post_form(
@@ -405,11 +421,14 @@ module TestNetHTTP_version_1_2_methods
   def test_request
     start {|http|
       _test_request__GET http
+      _test_request__accept_encoding http
       _test_request__file http
       # _test_request__range http   # WEBrick does not support Range: header.
       _test_request__HEAD http
       _test_request__POST http
       _test_request__stream_body http
+      _test_request__uri http
+      _test_request__uri_host http
     }
   end
 
@@ -424,6 +443,24 @@ module TestNetHTTP_version_1_2_methods
       end
       assert_equal $test_net_http_data.size, res.body.size
       assert_equal $test_net_http_data, res.body
+
+      assert res.decode_content, 'Bug #7831' if Net::HTTP::HAVE_ZLIB
+    }
+  end
+
+  def _test_request__accept_encoding(http)
+    req = Net::HTTP::Get.new('/', 'accept-encoding' => 'deflate')
+    http.request(req) {|res|
+      assert_kind_of Net::HTTPResponse, res
+      assert_kind_of String, res.body
+      unless self.is_a?(TestNetHTTP_v1_2_chunked)
+        assert_not_nil res['content-length']
+        assert_equal $test_net_http_data.size, res['content-length'].to_i
+      end
+      assert_equal $test_net_http_data.size, res.body.size
+      assert_equal $test_net_http_data, res.body
+
+      refute res.decode_content, 'Bug #7831' if Net::HTTP::HAVE_ZLIB
     }
   end
 
@@ -489,6 +526,51 @@ module TestNetHTTP_version_1_2_methods
     assert_equal data, res.body
   end
 
+  def _test_request__path(http)
+    uri = URI 'https://example/'
+    req = Net::HTTP::Get.new('/')
+
+    res = http.request(req)
+
+    assert_kind_of URI::Generic, req.uri
+
+    refute_equal uri, req.uri
+
+    assert_equal uri, res.uri
+
+    refute_same uri,     req.uri
+    refute_same req.uri, res.uri
+  end
+
+  def _test_request__uri(http)
+    uri = URI 'https://example/'
+    req = Net::HTTP::Get.new(uri)
+
+    res = http.request(req)
+
+    assert_kind_of URI::Generic, req.uri
+
+    refute_equal uri, req.uri
+
+    assert_equal req.uri, res.uri
+
+    refute_same uri,     req.uri
+    refute_same req.uri, res.uri
+  end
+
+  def _test_request__uri_host(http)
+    uri = URI 'http://example/'
+
+    req = Net::HTTP::Get.new(uri)
+    req['host'] = 'other.example'
+
+    res = http.request(req)
+
+    assert_kind_of URI::Generic, req.uri
+
+    assert_equal URI("http://example:#{http.port}"), res.uri
+  end
+
   def test_send_request
     start {|http|
       _test_send_request__GET http
@@ -517,15 +599,15 @@ module TestNetHTTP_version_1_2_methods
 
   def test_set_form
     require 'tempfile'
-    file = Tempfile.new('ruby-test')
-    file << "\u{30c7}\u{30fc}\u{30bf}"
-    data = [
-      ['name', 'Gonbei Nanashi'],
-      ['name', "\u{540d}\u{7121}\u{3057}\u{306e}\u{6a29}\u{5175}\u{885b}"],
-      ['s"i\o', StringIO.new("\u{3042 3044 4e9c 925b}")],
-      ["file", file, filename: "ruby-test"]
-    ]
-    expected = <<"__EOM__".gsub(/\n/, "\r\n")
+    Tempfile.create('ruby-test') {|file|
+      file << "\u{30c7}\u{30fc}\u{30bf}"
+      data = [
+        ['name', 'Gonbei Nanashi'],
+        ['name', "\u{540d}\u{7121}\u{3057}\u{306e}\u{6a29}\u{5175}\u{885b}"],
+        ['s"i\o', StringIO.new("\u{3042 3044 4e9c 925b}")],
+        ["file", file, filename: "ruby-test"]
+      ]
+      expected = <<"__EOM__".gsub(/\n/, "\r\n")
 --<boundary>
 Content-Disposition: form-data; name="name"
 
@@ -545,13 +627,12 @@ Content-Type: application/octet-stream
 \xE3\x83\x87\xE3\x83\xBC\xE3\x82\xBF
 --<boundary>--
 __EOM__
-    start {|http|
-      _test_set_form_urlencoded(http, data.reject{|k,v|!v.is_a?(String)})
-      _test_set_form_multipart(http, false, data, expected)
-      _test_set_form_multipart(http, true, data, expected)
+      start {|http|
+        _test_set_form_urlencoded(http, data.reject{|k,v|!v.is_a?(String)})
+        _test_set_form_multipart(http, false, data, expected)
+        _test_set_form_multipart(http, true, data, expected)
+      }
     }
-  ensure
-    file.close! if file
   end
 
   def _test_set_form_urlencoded(http, data)
@@ -576,12 +657,12 @@ __EOM__
 
   def test_set_form_with_file
     require 'tempfile'
-    file = Tempfile.new('ruby-test')
-    file.binmode
-    file << $test_net_http_data
-    filename = File.basename(file.to_path)
-    data = [['file', file]]
-    expected = <<"__EOM__".gsub(/\n/, "\r\n")
+    Tempfile.create('ruby-test') {|file|
+      file.binmode
+      file << $test_net_http_data
+      filename = File.basename(file.to_path)
+      data = [['file', file]]
+      expected = <<"__EOM__".gsub(/\n/, "\r\n")
 --<boundary>
 Content-Disposition: form-data; name="file"; filename="<filename>"
 Content-Type: application/octet-stream
@@ -589,31 +670,30 @@ Content-Type: application/octet-stream
 <data>
 --<boundary>--
 __EOM__
-    expected.sub!(/<filename>/, filename)
-    expected.sub!(/<data>/, $test_net_http_data)
-    start {|http|
-      data.each{|k,v|v.rewind rescue nil}
-      req = Net::HTTP::Post.new('/')
-      req.set_form(data, 'multipart/form-data')
-      res = http.request req
-      body = res.body
-      header, _ = body.split(/\r\n\r\n/, 2)
-      assert_match(/\A--(?<boundary>\S+)/, body)
-      /\A--(?<boundary>\S+)/ =~ body
-      expected = expected.gsub(/<boundary>/, boundary)
-      assert_match(/^--(?<boundary>\S+)\r\n/, header)
-      assert_match(
-        /^Content-Disposition: form-data; name="file"; filename="#{filename}"\r\n/,
-        header)
-      assert_equal(expected, body)
+      expected.sub!(/<filename>/, filename)
+      expected.sub!(/<data>/, $test_net_http_data)
+      start {|http|
+        data.each{|k,v|v.rewind rescue nil}
+        req = Net::HTTP::Post.new('/')
+        req.set_form(data, 'multipart/form-data')
+        res = http.request req
+        body = res.body
+        header, _ = body.split(/\r\n\r\n/, 2)
+        assert_match(/\A--(?<boundary>\S+)/, body)
+        /\A--(?<boundary>\S+)/ =~ body
+        expected = expected.gsub(/<boundary>/, boundary)
+        assert_match(/^--(?<boundary>\S+)\r\n/, header)
+        assert_match(
+          /^Content-Disposition: form-data; name="file"; filename="#{filename}"\r\n/,
+          header)
+        assert_equal(expected, body)
 
-      data.each{|k,v|v.rewind rescue nil}
-      req['Transfer-Encoding'] = 'chunked'
-      res = http.request req
-      #assert_equal(expected, res.body)
+        data.each{|k,v|v.rewind rescue nil}
+        req['Transfer-Encoding'] = 'chunked'
+        res = http.request req
+        #assert_equal(expected, res.body)
+      }
     }
-  ensure
-    file.close! if file
   end
 end
 
@@ -830,7 +910,9 @@ class TestNetHTTPLocalBind < Test::Unit::TestCase
 
     http = Net::HTTP.new(config('host'), config('port'))
     http.local_host = Addrinfo.tcp(config('host'), config('port')).ip_address
-    http.local_port = [*10000..20000].shuffle.first.to_s
+    http.local_port = Addrinfo.tcp(config('host'), 0).bind {|s|
+      s.local_address.ip_port.to_s
+    }
     assert_not_nil(http.local_host)
     assert_not_nil(http.local_port)
 
@@ -838,3 +920,4 @@ class TestNetHTTPLocalBind < Test::Unit::TestCase
     assert_equal(http.local_port, res.body)
   end
 end
+
