@@ -44,10 +44,10 @@ class TestArgf < Test::Unit::TestCase
     t
   end
 
-  def ruby(*args)
+  def ruby(*args, external_encoding: Encoding::UTF_8)
     args = ['-e', '$>.write($<.read)'] if args.empty?
     ruby = EnvUtil.rubybin
-    f = IO.popen([ruby] + args, 'r+')
+    f = IO.popen([ruby] + args, 'r+', external_encoding: external_encoding)
     yield(f)
   ensure
     f.close unless !f || f.closed?
@@ -247,8 +247,6 @@ class TestArgf < Test::Unit::TestCase
   end
 
   def test_inplace_stdin
-    t = make_tempfile
-
     assert_in_out_err(["-", "-"], <<-INPUT, [], /Can't do inplace edit for stdio; skipping/)
       ARGF.inplace_mode = '.bak'
       f = ARGF.dup
@@ -259,8 +257,6 @@ class TestArgf < Test::Unit::TestCase
   end
 
   def test_inplace_stdin2
-    t = make_tempfile
-
     assert_in_out_err(["-"], <<-INPUT, [], /Can't do inplace edit for stdio/)
       ARGF.inplace_mode = '.bak'
       while line = ARGF.gets
@@ -683,6 +679,49 @@ class TestArgf < Test::Unit::TestCase
     end
   end
 
+  def test_skip_in_each_line
+    ruby('-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      ARGF.each_line {|l| print l; ARGF.skip}
+    SRC
+      assert_equal("1\n3\n5\n", f.read, '[ruby-list:49185]')
+    end
+    ruby('-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      ARGF.each_line {|l| ARGF.skip; puts [l, ARGF.gets].map {|s| s ? s.chomp : s.inspect}.join("+")}
+    SRC
+      assert_equal("1+3\n4+5\n6+nil\n", f.read, '[ruby-list:49185]')
+    end
+  end
+
+  def test_skip_in_each_byte
+    ruby('-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      ARGF.each_byte {|l| print l; ARGF.skip}
+    SRC
+      assert_equal("135".unpack("C*").join(""), f.read, '[ruby-list:49185]')
+    end
+  end
+
+  def test_skip_in_each_char
+    [[@t1, "\u{3042}"], [@t2, "\u{3044}"], [@t3, "\u{3046}"]].each do |f, s|
+      File.write(f.path, s, mode: "w:utf-8")
+    end
+    ruby('-Eutf-8', '-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      ARGF.each_char {|l| print l; ARGF.skip}
+    SRC
+      assert_equal("\u{3042 3044 3046}", f.read, '[ruby-list:49185]')
+    end
+  end
+
+  def test_skip_in_each_codepoint
+    [[@t1, "\u{3042}"], [@t2, "\u{3044}"], [@t3, "\u{3046}"]].each do |f, s|
+      File.write(f.path, s, mode: "w:utf-8")
+    end
+    ruby('-Eutf-8', '-Eutf-8', '-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      ARGF.each_codepoint {|l| printf "%x:", l; ARGF.skip}
+    SRC
+      assert_equal("3042:3044:3046:", f.read, '[ruby-list:49185]')
+    end
+  end
+
   def test_close
     ruby('-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
       ARGF.close
@@ -770,4 +809,46 @@ class TestArgf < Test::Unit::TestCase
     bug5952 = '[ruby-dev:45160]'
     assert_ruby_status(["-e", "2.times {STDIN.tty?; readlines}"], "", bug5952)
   end
+
+  def test_lines
+    ruby('-W1', '-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      $stderr = $stdout
+      s = []
+      ARGF.lines {|l| s << l }
+      p s
+    SRC
+      assert_match(/deprecated/, f.gets)
+      assert_equal("[\"1\\n\", \"2\\n\", \"3\\n\", \"4\\n\", \"5\\n\", \"6\\n\"]\n", f.read)
+    end
+  end
+
+  def test_bytes
+    ruby('-W1', '-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      $stderr = $stdout
+      print Marshal.dump(ARGF.bytes.to_a)
+    SRC
+      assert_match(/deprecated/, f.gets)
+      assert_equal([49, 10, 50, 10, 51, 10, 52, 10, 53, 10, 54, 10], Marshal.load(f.read))
+    end
+  end
+
+  def test_chars
+    ruby('-W1', '-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      $stderr = $stdout
+      print [Marshal.dump(ARGF.chars.to_a)].pack('m')
+    SRC
+    assert_match(/deprecated/, f.gets)
+    assert_equal(["1", "\n", "2", "\n", "3", "\n", "4", "\n", "5", "\n", "6", "\n"], Marshal.load(f.read.unpack('m').first))
+    end
+  end
+
+  def test_codepoints
+    ruby('-W1', '-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
+      $stderr = $stdout
+      print Marshal.dump(ARGF.codepoints.to_a)
+    SRC
+      assert_match(/deprecated/, f.gets)
+      assert_equal([49, 10, 50, 10, 51, 10, 52, 10, 53, 10, 54, 10], Marshal.load(f.read))
+    end
+  end
 end
